/*
 *	$Id: adc.c,v 0.5 1996/11/07 14:09:49 glenn Exp glenn $
 */

/* Use this so we will be compiled with the same settings for
   CONFIG_MODVERSIONS as the kernel */
#include <linux/config.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/dma.h>

#include <fcntl.h>

#include "labpc.h"
#include "adc.h"
#include "counter.h"

#define MAX_DMA_SIZE	(64 * 1024)
#define MAX_COUNT	(64 * 1024)
#define	MAXBUFS		100

#ifndef ADCBUFSIZE
#define ADCBUFSIZE	1024
#endif
#ifndef ADCNBUF
#define ADCNBUF		2
#endif

int adcbufsize = ADCBUFSIZE;
int adcnbuf = ADCNBUF;

/* Clear the ADC - write to clear register and read 2 values from FIFO */
#define ClearADC( adc )			\
{ 					\
	outb_p( 0x00, (adc)->clear );	\
	inb_p( (adc)->fifo );		\
	inb_p( (adc)->fifo );		\
}

/* Start an ADC conversion by writing to the ADC start register */
#define StartADCConversion( adc )	\
	outb_p( 0x00, (adc)->start )

/* Read a 16-bit ADC conversion value from the ADC FIFO */
#define ReadADCFIFO( adc )		\
	(inb_p( (adc)->fifo ) | (inb_p( (adc)->fifo ) << 8))

/* Accessing Board command and status registers */
#define ADCstatus( adc )		\
	(readport( &(((adc)->board)->regs.stat) ))

/* Enable ADC DMA interrupt generation */
#define EnableDMAInterrupts( adc )	\
	setbits( &((adc)->board->regs.cmd3), DMAEN | TCINTEN | ERRINTEN )

/* Clear an ADC DMA Terminal Count interrupt */
#define ClearDMAInterrupt( adc )	\
	outb_p( 0x00, (adc)->dmaintclr )

/* Setup the interval mode control registers */
#define WriteIntervalCounter( adc, count )	\
	outb_p( count, (adc)->intvlcntr );	\
	outb_p( 1, (adc)->intvlstrobe )

/* Enable interrupts on ADC FIFO */
#define EnableADCInterrupts( adc )	\
	setbits( &((adc)->board->regs.cmd3), FIFOINTEN | ERRINTEN )

/* Disable all ADC interrupt generation */
#define DisableADCInterrupts( adc )		\
	clearbits( &((adc)->board->regs.cmd3),	\
		   DMAEN | TCINTEN | ERRINTEN | FIFOINTEN )

/* Static functions don't show up in ksyms - make external for debugging */
#ifndef STATIC
#define STATIC
#endif /* STATIC */

STATIC void free_adc_buffers( ADC_t *adc )
{
	DmaBuffer_t *p;

	adc->head = adc->tail = adc->buflist;
	if (adc->buflist == NULL)
		return;
	for (p = adc->buflist; p < adc->buflist + adc->nbuf; p++) {
		if (p->buffer)
			kfree( p->buffer );
	}
	kfree( adc->buflist );
	adc->head = adc->tail = adc->buflist = NULL;

	return;
}

STATIC int reset_adc_buffers( ADC_t *adc )
{
	DmaBuffer_t	*p;

	adc->head = adc->tail = adc->buflist;
	if (adc->buflist == NULL)
		return 0;
	for (p = adc->buflist; p < adc->buflist + adc->nbuf; p++) {
		p->full = p->count = 0;
		p->fillp = p->bufp = p->buffer;
	}

	return 0;
}

STATIC int allocate_adc_buffers( ADC_t *adc )
{
	DmaBuffer_t	*p;

	if (adc->bufsize * sizeof( short ) > MAX_DMA_SIZE) {
		printk( "LabPC+: Requested DMA buffer size, %d, too large\n",
		       adc->bufsize );
		adc->bufsize = MAX_DMA_SIZE / sizeof( short );
		printk( "LabPC+: Using DMA buffer size, %d\n",
		       adc->bufsize );
	}

	adc->buflist = (DmaBuffer_t *)kmalloc( adc->nbuf *
					       sizeof( DmaBuffer_t ),
					       GFP_KERNEL );
	adc->head = adc->tail = adc->buflist;
	if (adc->buflist == NULL) {
		printk( "LabPC+: ADC unable to allocate DMA buffer record\n" );
		return -(ENOMEM);
	}
	for (p = adc->buflist; p < adc->buflist + adc->nbuf; p++) {
		p->size = adc->bufsize;
		p->buffer = NULL;
	}
	for (p = adc->buflist; p < adc->buflist + adc->nbuf; p++) {
		p->buffer = (short *)kmalloc( p->size * sizeof( short ),
					      GFP_KERNEL | GFP_DMA );
		if (p->buffer == NULL) {
			printk( "LabPC+: ADC unable to allocate DMA buffer\n" );
			free_adc_buffers( adc );
			return -(ENOMEM);
		}
	}
	reset_adc_buffers( adc );

	return 0;
}

/* Stop data collection and reset the hardware - but don't reset the
 * status flags - ResetADC() does that */
STATIC void StopADC( ADC_t *adc )
{
	DisableADCInterrupts( adc );
	adc->waitfortrig = 0;

	if (adc->dmaen && adc->board->dma) {
		disable_dma( adc->board->dma );
		if (adc->owndma) {
			free_dma( adc->board->dma );
			adc->owndma = 0;
		}
	}
	reset_adc_buffers( adc );

	/* Clear the trigger condition */
	clearbits( &(adc->board->regs.cmd2), SWTRIG | HWTRIG | PRETRIG );
	clearbits( &(adc->board->regs.cmd4), EOIRCV | INTSCAN );

	ClearADC( adc );
	adc->residue = adc->count = 0;
	adc->status |= ADC_STATUS_STOPPED;
	/* Wake up any processes which may be waiting for a read */
	/* NO - is reponsibility of the calling function - depends on context */
	/* ??? Not possible - only one process at a time can read() - but
           what about select() ??? */
	/* wake_up_interruptible( &(adc->wait) ); */
	/* wake_up_interruptible( &(adc->exwait) ); */
}

/* ResetADC() stops the ADC and sets status to inactive */
STATIC void ResetADC( ADC_t *adc )
{
	if (!(adc->status & ADC_STATUS_STOPPED) &&
	    (adc->status != ADC_STATUS_INACTIVE)) {
		StopADC( adc );
	}
	adc->status = ADC_STATUS_INACTIVE;
}

STATIC int ConfigureADCTiming( ADC_t *adc, int count )
{
	if (count > 1) {
		/* This follows the recommended sequence in
		 * Chapter 5 of the Lab-PC+ manual
		 * ("Programming Multiple A/D Conversions with
		 * Channel Scanning"). I clear the bit twice
		 * just to be sure - this may not be
		 * necessary.  */
		Counter_t *counter;

		/* Program the Channel Number & Scanning mode */
		clearbits( &(adc->board->regs.cmd1), SCANEN );
		clearbits( &(adc->board->regs.cmd1), SCANEN );
		if (adc->scan) {
			/* Channel Scanning is ON */
			if (adc->intscan) {
				/* Also Interval Scanning is ON */
				setbits( &(adc->board->regs.cmd4), INTSCAN );
				if (adc->intvlclock == 0) {
					/* Extern. Gate for Interval Scanning */
					setbits( &(adc->board->regs.cmd4),
						 EOIRCV );
				}
			} else {
				setbits( &(adc->board->regs.cmd1), SCANEN );
			}
		} else if (adc->intscan) {
			/* Single Channel Interval Acquisition Mode */
			WriteIntervalCounter( adc, adc->intvlcount );
			setbits( &(adc->board->regs.cmd4), INTSCAN );
			if (adc->intvlclock == 0) {
				/* Extern. Gate for Interval Acquisition Mode */
				setbits( &(adc->board->regs.cmd4), EOIRCV );
			}
		}

		/* Program the Sample timing clock */
		counter = FINDCOUNTER( adc->board, DAQ_COUNTERA1 );
		SetHigh8253( counter );
		if (adc->clock) {
			counter = FINDCOUNTER( adc->board, DAQ_COUNTERA0 );
			SetMode8253( counter, MODE_RATE_GENERATOR );
			Write8253( counter, adc->clock );

			if (adc->timebase != 2) {
				/* Use counterB0 as the clock timebase */
				counter = FINDCOUNTER( adc->board, DAQ_COUNTERB0 );
				SetMode8253( counter, MODE_SQUARE_WAVE );
				Write8253( counter, adc->timebase );
			}
		} else {
			/* Use EXTCONV* for external ADC trigger */
			clearbits( &(adc->board->regs.cmd4), ECKDRV );
			counter = FINDCOUNTER( adc->board, DAQ_COUNTERA0 );
			SetHigh8253( counter );
		}
		ClearADC( adc );

		/* Program the Sample counter */
		counter = FINDCOUNTER( adc->board, DAQ_COUNTERA1 );
		if (count <= MAX_COUNT) {
			SetMode8253( counter, MODE_TERMINAL_COUNT );
			Write8253( counter, count - 2 );
		} else {
			SetLow8253( counter );	/* Freerun mode */
		}
	}

	return 0;
}

STATIC int StartADC( ADC_t *adc, int count )
{
	if (adc->head == NULL) {
		printk( "LabPC+: adc: No buffers for data acquisition.\n" );
		return -(EIO);
	}
	adc->status = ADC_STATUS_STARTING;
	adc->residue = adc->count = count;
	reset_adc_buffers( adc );
	adc->head->count = (adc->residue > adc->head->size) ? adc->head->size : adc->residue;
	if (adc->dmaen) {
		/* Setup the DMA transfers */
		if (!adc->board->dma) {
			printk( "LabPC+: Attempted DMA with no DMA channel\n" );
			ResetADC( adc );
			return -(EINVAL);
		}
		if (!adc->owndma) {
			if (request_dma( adc->board->dma, "labpc" )) {
				printk( "LabPC+: can't request DMA channel %d\n", adc->board->dma );
				ResetADC( adc );
				return -(EIO);
			}
			adc->owndma = 1;
		}
		disable_dma( adc->board->dma );
		EnableDMAInterrupts( adc );
		clear_dma_ff( adc->board->dma );
		set_dma_addr( adc->board->dma,
			      (unsigned long)(void *)adc->head->buffer );
		set_dma_count( adc->board->dma, adc->head->count * sizeof( short ) );
		set_dma_mode( adc->board->dma, DMA_MODE_READ );
		enable_dma( adc->board->dma );
	} else {
		EnableADCInterrupts( adc );
	}

	if (adc->intscan) {
		/* Program the interval counter source */
		Counter_t *counter;

		counter = FINDCOUNTER( adc->board, DAQ_COUNTERB1 );
		if (adc->intvlclock > 0) {
			SetMode8253( counter, MODE_RATE_GENERATOR );
			Write8253( counter, adc->intvlclock );
		}
		if (adc->scan)
			setbits( &(adc->board->regs.cmd1), SCANEN );
	}

	switch (adc->triggermode) {
	case ADC_AUTOTRIGGER:
		/* Trigger the data acquisition sequence */
		if (count > 1) {
			setbits( &(adc->board->regs.cmd2), SWTRIG );
		} else {
			StartADCConversion( adc );
		}
		break;
	case ADC_SOFTTRIGGER:
		/* Wait for software ADC trigger */
		adc->waitfortrig = 1;
		break;
	case ADC_HARDTRIGGER:
		/* Wait for hardware ADC trigger */
		break;
	case ADC_PRETRIGGER:
		dprintk(( "ADC: trigger type not implemented\n" ));
		ResetADC( adc );
		return -(EINVAL);
		break;
	default:
		dprintk(( "ADC: unknown trigger type: %d\n",
			  adc->triggermode ));
		ResetADC( adc );
		return -(EINVAL);
		break;
	}

	adc->status = ADC_STATUS_ACTIVE;
	return 0;
}

STATIC int read_adc( struct inode *inode, struct file *file,
		     char *buffer, int count )
{
	ADC_t	*adc;
	int 	retval = 0;
	int	n;

	if ((adc = GETADC( inode->i_rdev )) == NULL)
		return -(EINVAL);

	count /= sizeof( short );

	if (count < 1)
		return -(EINVAL);

	/* Lockout other processes from accessing the ADC */
	down( &(adc->sem) );

	if (adc->status == ADC_STATUS_INACTIVE) {
		if (adc->autoacquire) {
			/* Start a data acquisition sequence */
			ResetADC( adc );
			if ((retval = ConfigureADCTiming( adc, count ))) {
				up( &(adc->sem) );
				return retval;
			}
			cli();
			if ((retval = StartADC( adc, count ))) {
				sti();
				up( &(adc->sem) );
				return retval;
			}
		}
		if (file->f_flags & O_NONBLOCK) {
			up( &(adc->sem) );
			return -(EWOULDBLOCK);
		}
	}
	if (adc->head == NULL) {
		sti();
		printk( "LabPC+: adc: No data buffers for data acquisition.\n" );
		up( &(adc->sem) );
		return -(EIO);
	}

	/* Read data from the ADC data buffers */
	while (!(adc->status & ADC_STATUS_ERROR)) {
		while (adc->tail->full) {
			/* There's data in the buffer - get it :-) */
			sti();
			n = (adc->tail->count < count) ? adc->tail->count : count;
			memcpy_tofs( buffer, adc->tail->bufp, n * sizeof( short ) );
			buffer += n * sizeof( short );
			retval += n * sizeof( short );
			count -= n;
			adc->count -= n;
			adc->tail->count -= n;
			if (adc->tail->count == 0) {
				adc->tail->full = 0;
				if (++adc->tail >= adc->buflist + adc->nbuf)
					adc->tail = adc->buflist;
				adc->tail->bufp = adc->tail->buffer;
			} else {
				adc->tail->bufp += n;
			}
			if (adc->count == 0) {
				/* Acquisition complete - reset the ADC */
				ResetADC( adc );
				up( &(adc->sem) );
				return retval;
			} else if (count == 0) {
				/* This read() is complete */
				up( &(adc->sem) );
				return retval;
			}
			cli();
		}
		if (file->f_flags & O_NONBLOCK) {
			/* Non-blocking - return immediately */
			sti();
			up( &(adc->sem) );
			if (retval == 0)
				return -(EWOULDBLOCK);
			return retval;
		}
		/* Block and wait for more data */
		interruptible_sleep_on( &(adc->wait) );
		if (current->signal & ~current->blocked) {
			ResetADC( adc );
			up( &(adc->sem) );
			dprintk(( "ADC: interrupted\n" ));
			return -(EINTR);
		}
		cli();
	}
	sti();

	if (adc->status & ADC_STATUS_ERROR) {
		if (adc->status & ADC_STATUS_OVERRUN) {
			dprintregs( adc->board );
			dprintk(( "LabPC+: ADC overrun detected\n" ));
			ResetADC( adc );
			retval = -(EIO);
		} else if (adc->status & ADC_STATUS_OVERFLOW) {
			dprintregs( adc->board );
			dprintk(( "LabPC+: ADC overflow detected\n" ));
			ResetADC( adc );
			retval = -(EIO);
		} else if (adc->status & ADC_STATUS_BUFFER_OVERFLOW) {
			dprintregs( adc->board );
			dprintk(( "LabPC+: ADC buffer overflow detected\n" ));
			ResetADC( adc );
			retval = -(EIO);
		} else if (adc->status & ADC_STATUS_IO_ERROR) {
			dprintregs( adc->board );
			dprintk(( "LabPC+: ADC IO error detected\n" ));
			ResetADC( adc );
			retval = -(EIO);
		} else {
			dprintregs( adc->board );
			dprintk(( "LabPC+: ADC unknown error detected\n" ));
			ResetADC( adc );
			retval = -(EIO);
		}
	} else if (adc->status & ADC_STATUS_STOPPED) {
		/* For a stopped ADC return data we have so far */
		dprintregs( adc->board );
		dprintk(( "LabPC+: ADC has been stopped\n" ));
		ResetADC( adc );
	}

	/* Release the exclusion semaphore */
	up( &(adc->sem) );

	return retval;
}

STATIC int select_adc( struct inode *inode, struct file *file,
		       int sel_type, select_table *wait )
{
	ADC_t	*adc;

	/* Which ADC? */
	if ((adc = GETADC( inode->i_rdev )) == NULL)
		return -(EINVAL);

	if (sel_type == SEL_IN) {
		cli();
		if ((adc->status & ADC_STATUS_ACTIVE) && adc->tail->full) {
			sti();
			return 1;
		}
		select_wait( &(adc->wait), wait );
	} else if (sel_type == SEL_EX) {
		cli();
		if (adc->status & (ADC_STATUS_STOPPED | ADC_STATUS_ERROR)) {
			sti();
			return 1;
		}
		select_wait( &(adc->exwait), wait );
	} else {
		return -(ENOSYS);
	}
	sti();

	return 0;
}

STATIC int init_adc( ADC_t *adc )
{
	static struct semaphore mutex = MUTEX;

	dprintregs( adc->board );
	DisableADCInterrupts( adc );
	ClearADC( adc );

	/* Initialise the ADC registers */

	/* ADC channel 0: ADC output -2048 to 2047 : Gain = 0 : Disable Scan */
	clearbits( &(adc->board->regs.cmd1), (MA0 | MA1 | MA2 | TWOSCMP |
				     GAIN0 | GAIN1 | GAIN2 | SCANEN) );
	setbits( &(adc->board->regs.cmd1), TWOSCMP );
	/* Disable pretrigger: Disable Hardware trigger :
	 * Disable Software trigger: Time Base = 0 */
	clearbits( &(adc->board->regs.cmd2), (PRETRIG | HWTRIG | SWTRIG) );
	/* Disable Interval scanning: Disable clock for interval scanning:
	 * Direction of EXTCONV : Single ended ADC input */
	clearbits( &(adc->board->regs.cmd4), (INTSCAN | EOIRCV | ECKDRV | SE_D) );

	/* Set the data ports for the ADC */
	adc->fifo	= adc->board->base + DAQ_FIFO;
	adc->clear	= adc->board->base + DAQ_ADCCLR;
	adc->start	= adc->board->base + DAQ_STRTCVT;
	adc->dmaintclr	= adc->board->base + DAQ_DMATCCLR;
	adc->intvlcntr	= adc->board->base + DAQ_INTVLDATA;
	adc->intvlstrobe = adc->board->base + DAQ_INTVLSTROBE;

	/* Initialise options and state variables */
	adc->scan	= 0;
	adc->gain	= ADC_GAIN_1;
	adc->polarity	= ADC_BIPOLAR;
	adc->channel	= 0;
	adc->timebase	= 2;
	adc->autoacquire= 1;
	adc->triggermode= ADC_AUTOTRIGGER;
	adc->inputmode	= ADC_SINGLEENDED;
	adc->eckdrv	= 0;
	adc->intscan	= 0;
	adc->intvlclock	= 0;
	adc->intvlcount	= 0;
	adc->clock	= 1000;
	adc->waitfortrig= 0;
	adc->dmaen	= (adc->board->dma != 0);
	adc->owndma	= 0;

	/* Access control variables */
	adc->sem   	= mutex;
	adc->wait  	= NULL;
	adc->exwait  	= NULL;
	adc->status	= ADC_STATUS_INACTIVE;

	dprintregs( adc->board );

	return 0;
}

STATIC int ioctl_adc( struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg )
{
	ADC_t 	*adc;
	int	stat = 0;
	
	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	/* Which ADC? */
	if ((adc = GETADC( inode->i_rdev )) == NULL)
		return -(EINVAL);

	switch (cmd) {
	case ADC_INIT:
		down( &(adc->sem) );
		if (adc->owndma) {
			disable_dma( adc->board->dma );
			free_dma( adc->board->dma );
			adc->owndma = 0;
		}
		init_adc( adc );
		up( &(adc->sem) );
		break;
	case ADC_CLEAR:
		/* Clear the ADC fifo buffer */
		down( &(adc->sem) );
		ClearADC( adc );
		up( &(adc->sem) );
		break;
	case ADC_TRIGGER:
		/* Trigger an ADC conversion sequence */
		if (adc->triggermode == ADC_SOFTTRIGGER && adc->waitfortrig) {
			if (adc->count > 1) {
				setbits( &(adc->board->regs.cmd2), SWTRIG );
			} else {
				StartADCConversion( adc );
			}
		} else {
			stat = -(EINVAL);
		}
		break;
	case ADC_ACQUIRE:
		/* Start an ADC acquisition sequence */
		ResetADC( adc );
		if (!((stat = ConfigureADCTiming( adc, arg )))) {
			cli();
			stat = StartADC( adc, arg );
			sti();
		}
		break;
	case ADC_STOP:
		/* Stop an ADC acquisition sequence */
		ResetADC( adc );
		wake_up_interruptible( &(adc->wait) );
		wake_up_interruptible( &(adc->exwait) );
		break;
	case ADC_GET_STATUS:
		/* Read the ADC status word */
		put_fs_long( adc->status, (unsigned long *)arg );
		break;
	case ADC_SET_SCAN:
		/* Enable/disable ADC input channel scanning */
		down( &(adc->sem) );
		adc->scan = arg;
		up( &(adc->sem) );
		break;
	case ADC_GET_SCAN:
		/* Get the DAC clock counter value */
		put_fs_long( adc->scan, (unsigned long *)arg );
		break;
	case ADC_SET_GAIN:
		/* Set the ADC gain */
		down( &(adc->sem) );
		if (arg > ADC_GAIN_MAX) {
			stat = -(EINVAL);
		} else {
			writebits( &(adc->board->regs.cmd1),
				   (arg << 4) & (GAIN0 | GAIN1 | GAIN2),
				   (~(arg << 4)) & (GAIN0 | GAIN1 | GAIN2) );
			adc->gain = arg;
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_GAIN:
		/* Get the DAC clock counter value */
		put_fs_long( adc->gain, (unsigned long *)arg );
		break;
	case ADC_SET_POLARITY:
		down( &(adc->sem) );
		if (arg == ADC_BIPOLAR) {
			/* Signal is bipolar - read data in 2's complement */
			setbits( &(adc->board->regs.cmd1), TWOSCMP );
			adc->polarity = arg;
		} else if (arg == ADC_UNIPOLAR) {
			/* Signal is unipolar - read data in unsigned binary */
			clearbits( &(adc->board->regs.cmd1), TWOSCMP );
			adc->polarity = arg;
		} else {
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_POLARITY:
		/* Get the DAC clock counter value */
		put_fs_long( adc->polarity, (unsigned long *)arg );
		break;
	case ADC_SET_CHANNEL:
		/* ADC input channel no. OR max channel for scanning */
		down( &(adc->sem) );
		if (arg >= ADC_NUM_CHANNELS) {
			stat = -(EINVAL);
		} else {
			writebits( &(adc->board->regs.cmd1),
				  arg & (MA0 | MA1 | MA2),
				  (~arg) & (MA0 | MA1 | MA2) );
			adc->channel = arg;
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_CHANNEL:
		put_fs_long( adc->channel, (unsigned long *)arg );
		break;
	case ADC_SET_TIMEBASE:
		down( &(adc->sem) );
		if (arg == 2) {
			adc->timebase = arg;
			clearbits( &(adc->board->regs.cmd2), TBSEL );
		} else if (arg > 2) {
			adc->timebase = arg;
			setbits( &(adc->board->regs.cmd2), TBSEL );
		} else {
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_TIMEBASE:
		put_fs_long( adc->timebase, (unsigned long *)arg );
		break;
	case ADC_SET_TRIGGER_MODE:
		/* Set the triggering mode for data acquisition */
		down( &(adc->sem) );
		switch (arg) {
		case ADC_SOFTTRIGGER:
			adc->triggermode = arg;
			clearbits( &(adc->board->regs.cmd2),
				   (SWTRIG | HWTRIG | PRETRIG) );
			break;
		case ADC_HARDTRIGGER:
			adc->triggermode = arg;
			writebits( &(adc->board->regs.cmd2),
				   HWTRIG, (SWTRIG | PRETRIG) );
			break;
		case ADC_PRETRIGGER:
			adc->triggermode = arg;
			writebits( &(adc->board->regs.cmd2),
				   PRETRIG, (SWTRIG | HWTRIG) );
			break;
		case ADC_AUTOTRIGGER:
			adc->triggermode = arg;
			clearbits( &(adc->board->regs.cmd2),
				   (PRETRIG | SWTRIG | HWTRIG) );
			break;
		default:
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_TRIGGER_MODE:
		/* Get the ADC trigger mode */
		put_fs_long( adc->triggermode, (unsigned long *)arg );
		break;
	case ADC_SET_CLOCK:
		adc->clock = arg;
		if (arg > 0) {
			stat = Write8253( FINDCOUNTER( adc->board,
						       DAQ_COUNTERA0 ), arg );
		} else {
			stat = SetHigh8253( FINDCOUNTER( adc->board,
							 DAQ_COUNTERA0 ) );
		}
		break;
	case ADC_GET_CLOCK:
		/* Get the ADC clock counter value */
		put_fs_long( adc->clock, (unsigned long *)arg );
		break;
	case ADC_SET_DMA:
		/* Enable DMA for data t/fers */
		down( &(adc->sem) );
		if (arg && adc->board->dma == 0)
			stat = -(EINVAL);
		else
			adc->dmaen = (arg != 0);
		up( &(adc->sem) );
		break;
	case ADC_GET_DMA:
		/* Get the ADC dma mode value */
		put_fs_long( adc->dmaen, (unsigned long *)arg );
		break;
	case ADC_SET_CONNECTION:
		/* ADC inputs are single ended/differential */
		down( &(adc->sem) );
		if (arg == ADC_DIFFERENTIAL) {
			adc->inputmode = arg;
			setbits( &(adc->board->regs.cmd4), SE_D );
		} else if (arg == ADC_SINGLEENDED) {
			adc->inputmode = arg;
			clearbits( &(adc->board->regs.cmd4), SE_D );
		} else {
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_CONNECTION:
		/* Get the electrical connection mode value */
		put_fs_long( adc->inputmode, (unsigned long *)arg );
		break;
	case ADC_SET_ECKDRV:
		/* Mode for the EXTCONV line on the I/O connector */
		down( &(adc->sem) );
		if ((adc->eckdrv = (arg != 0))) {
			setbits( &(adc->board->regs.cmd4), ECKDRV );
		} else {
			clearbits( &(adc->board->regs.cmd4), ECKDRV );
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_ECKDRV:
		/* Get the EXTCONV line mode value */
		put_fs_long( adc->eckdrv, (unsigned long *)arg );
		break;
	case ADC_SET_INTSCAN:
		/* Enable Interval Scanning */
		down( &(adc->sem) );
		adc->intscan = arg;
		up( &(adc->sem) );
		break;
	case ADC_GET_INTSCAN:
		/* Get interval scanning mode */
		put_fs_long( adc->intscan, (unsigned long *)arg );
		break;
	case ADC_SET_INTCOUNT:
		/* Enable Interval Scanning */
		down( &(adc->sem) );
		adc->intvlcount = arg;
		up( &(adc->sem) );
		break;
	case ADC_GET_INTCOUNT:
		/* Get interval scanning mode */
		put_fs_long( adc->intvlcount, (unsigned long *)arg );
		break;
	case ADC_SET_INTCLOCK:
		/* Set the interval clock value */
		down( &(adc->sem) );
		adc->intvlclock = arg;
		up( &(adc->sem) );
		break;
	case ADC_GET_INTCLOCK:
		/* Get interval clock value */
		put_fs_long( adc->intvlclock, (unsigned long *)arg );
		break;
	case ADC_SET_BUFSIZE:
		if (arg * sizeof( short ) > MAX_DMA_SIZE) {
			return -(EINVAL);
		} else {
			down( &(adc->sem) );
			free_adc_buffers( adc );
			adc->bufsize = arg;
			stat = allocate_adc_buffers( adc );
			up( &(adc->sem) );
		}
		break;
	case ADC_GET_BUFSIZE:
		put_fs_long( adc->bufsize, (unsigned long *)arg );
		break;
	case ADC_SET_NBUF:
		if (arg < 2 || arg > MAXBUFS) {
			return -(EINVAL);
		} else {
			down( &(adc->sem) );
			free_adc_buffers( adc );
			adc->nbuf = arg;
			stat = allocate_adc_buffers( adc );
			up( &(adc->sem) );
		}
		break;
	case ADC_GET_NBUF:
		put_fs_long( adc->nbuf, (unsigned long *)arg );
		break;
	case ADC_SET_AUTOACQUIRE:
		/* Enable DMA for data t/fers */
		down( &(adc->sem) );
		adc->autoacquire = (arg != 0);
		up( &(adc->sem) );
		break;
	case ADC_GET_AUTOACQUIRE:
		/* Get the ADC dma mode value */
		put_fs_long( adc->autoacquire, (unsigned long *)arg );
		break;
	default:
		return -(ENOSYS);
		break;
	}

	return stat;
}

STATIC int open_adc( struct inode *inode, struct file *file )
{
	ADC_t 	*adc;
	
	dprintk(( "LabPC+: open_adc()\n" ));
		
	/* Which ADC? */
	if ((adc = GETADC( inode->i_rdev )) == NULL)
		return -(EINVAL);

	adc->open++;

	return 0;
}

STATIC void release_adc(struct inode *inode, struct file *file)
{
	ADC_t 		*adc;

	dprintk(( "LabPC+: release_adc()\n" ));

	if ((adc = GETADC( inode->i_rdev )) == NULL)
		return;

	adc->open--;

	if (adc->open == 0) {
		/* Last close - shutdown the ADC if in running mode */
		if (ADCstatus( adc ) & (DAVAIL | ADC_OVERRUN | ADC_OVERFLOW))
			ResetADC( adc );
	}

	labpc_close( inode, file );
}

STATIC void abort_adc_xfer( ADC_t *adc, int status )
{
	if (status & ADC_OVERRUN)
		printk( "LabPC+: IRQ: ADC overrun\n" );
	if (status & ADC_OVERFLOW)
		printk( "LabPC+: IRQ: ADC overflow\n" );
	if (status & ADC_STATUS_IO_ERROR)
		printk( "LabPC+: IRQ: ADC DMATC bit set not in DMA mode!\n" );
	if (status & ADC_STATUS_BUFFER_OVERFLOW)
		printk( "LabPC+: IRQ: ADC data buffer overflow\n" );
	if (status & ADC_OVERFLOW)
		printk( "LabPC+: IRQ: ADC overflow\n" );

	dprintregs( adc->board );
	adc->status |= (status | ADC_STATUS_ACQUISITION_COMPLETE);
	DisableADCInterrupts( adc );
	if (adc->dmaen)
		ClearDMAInterrupt( adc );
	clearbits( &(adc->board->regs.cmd2),
		   SWTRIG | HWTRIG | PRETRIG );
	clearbits( &(adc->board->regs.cmd4),
		   EOIRCV | INTSCAN );
	ClearADC( adc );
	wake_up_interruptible( &(adc->exwait) );

	return;
}

void adc_interrupt( ADC_t *adc, int status )
{
	if (status & (ADC_OVERRUN | ADC_OVERFLOW)) {
		abort_adc_xfer( adc, status & (ADC_OVERRUN | ADC_OVERFLOW) );
	} else if (status & DMATC) {
		if (!adc->dmaen || !adc->head ||
		    !(adc->status & ADC_STATUS_ACTIVE)) {
			abort_adc_xfer( adc, ADC_STATUS_IO_ERROR );
			return;
		}
		adc->head->full = 1;
		adc->residue -= adc->head->count;

		/* Get the next data transfer buffer */
		if (++adc->head >= adc->buflist + adc->nbuf)
			adc->head = adc->buflist;
		if (adc->residue) {
			if (adc->head->full) {
				abort_adc_xfer( adc, ADC_STATUS_BUFFER_OVERFLOW );
				return;
			}
			/* setup next data transfer */
			adc->head->count = (adc->residue > adc->head->size) ? adc->head->size : adc->residue;

			set_dma_addr( adc->board->dma, (unsigned long)(void *)adc->head->buffer );
			set_dma_count( adc->board->dma, adc->head->count * sizeof( adc->head->buffer[0] ) );
			enable_dma( adc->board->dma );
		} else {
			DisableADCInterrupts( adc );
			adc->status |= ADC_STATUS_ACQUISITION_COMPLETE;
		}
		ClearDMAInterrupt( adc );

		wake_up_interruptible( &(adc->wait) );
	} else if (status & DAVAIL) {
		if (adc->dmaen)
			return;
		if (!adc->head || !(adc->status & ADC_STATUS_ACTIVE)) {
			abort_adc_xfer( adc, ADC_STATUS_IO_ERROR );
			return;
		}
		*(adc->head->fillp++) = ReadADCFIFO( adc );
		if (adc->head->fillp >= adc->head->buffer + adc->head->count) {
			adc->head->full = 1;
			adc->residue -= adc->head->count;
			/* Get the next data transfer buffer */
			if (++adc->head >= adc->buflist + adc->nbuf)
				adc->head = adc->buflist;
			/* setup next transfer */
			if (adc->residue) {
				/* Check the back buffer has been emptied */ 
				if (adc->head->full) {
					abort_adc_xfer( adc, ADC_STATUS_BUFFER_OVERFLOW );
					return;
				}

				adc->head->count = (adc->residue > adc->head->size) ? adc->head->size : adc->residue;
				adc->head->fillp = adc->head->buffer;
			} else {
				DisableADCInterrupts( adc );
				adc->status |= ADC_STATUS_ACQUISITION_COMPLETE;
			}

			wake_up_interruptible( &(adc->wait) );
		}
	} else {
		dprintk(( "Lab-PC+: IRQ: ADC interrupt for no good reason\n" ));
		dprintregs( adc->board );
	}

	return;
}

struct file_operations adc_fops = {
	NULL,			/* seek */
	read_adc,
	NULL,			/* write */
	NULL,			/* readdir */
	select_adc,
	ioctl_adc,		/* ioctl */
	NULL,			/* mmap */
	open_adc,
	release_adc,
};

/* Functions to load and unload the ADC structure */

ADC_t *get_adc( LabPC_t *board, int minor )
{
	ADC_t	*adc;
	
	adc = (ADC_t *)kmalloc( sizeof( ADC_t ), GFP_KERNEL );

	adc->board = board;
	adc->open = 0;
	init_adc( adc );

	adc->bufsize = adcbufsize;
	adc->buflist = NULL;
	adc->nbuf = adcnbuf;
	if (allocate_adc_buffers( adc ) < 0) {
		kfree( adc );
		return NULL;
	}
	adc->residue = adc->count = 0;

	return adc;
}

void free_adc( ADC_t *adc )
{
	if (adc != NULL) {
		if (adc->owndma) {
			disable_dma( adc->board->dma );
			free_dma( adc->board->dma );
			adc->owndma = 0;
		}
		free_adc_buffers( adc );
		init_adc( adc );
		kfree( adc );
	}
}

