/*
 *	$Id: counter.c,v 0.5 1996/11/07 14:09:49 glenn Exp glenn $
 */

#include <linux/config.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "counter.h"
#include "labpc.h"

/* Static functions don't show up in ksyms - make external for debugging */
#ifndef STATIC
#define STATIC static
#endif

STATIC int write_counter( struct inode *inode, struct file *file,
		        const char *buffer, int count )
{
	Counter_t	*counter;

	dprintk(( "LabPC+: write_counter()\n" ));
		
	/* Must write 16 bit int to counter */
	if (count != sizeof( ushort ))
		return -(EINVAL);

	/* Which counter? */
	if ((counter = GETCOUNTER( inode->i_rdev )) == NULL)
		return -(EINVAL);

	Write8253( counter, get_fs_word( buffer ) );

	return sizeof( short );
}

STATIC int read_counter( struct inode *inode, struct file *file,
			char *buffer, int count )
{
	Counter_t	*counter;
	
	dprintk(( "LabPC+: read_counter()\n" ));
		
	/* Must read 16 bit int to counter */
	if (count < sizeof( ushort ))
		return -(EINVAL);

	/* Which counter? */
	if ((counter = GETCOUNTER( inode->i_rdev )) == NULL)
		return -(EINVAL);

	put_fs_word( Read8253( counter ), (short *)buffer );

	return sizeof( short );
}

STATIC int ioctl_counter( struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg )
{
	Counter_t	*counter;
	int		stat;

	dprintk(( "LabPC+: ioctl_counter()\n" ));
		
	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	/* Which counter? */
	if ((counter = GETCOUNTER( inode->i_rdev )) == NULL)
		return -(EINVAL);

	switch (cmd) {
	case COUNTER_SET_MODE:
		/*
		 * Set the mode for this counter. This will not affect
		 * currently running counts. The new mode will be implemented
		 * on the next "write" to the counter
		 */
		/* Check we have a valid mode */
		if (arg >= NUMBER_OF_MODES)
			return -(EINVAL);	

		return SetMode8253( counter, arg );
		break;
	case COUNTER_GET_MODE:
		put_fs_long( counter->mode, (unsigned long *)arg );
		break;
	default:
		return -(EINVAL);
	}

	return 0;			/* NoERR */
}

STATIC int open_counter( struct inode *inode, struct file *file )
{
	int		minor;
	Counter_t	*counter;

	dprintk(( "LabPC+: open_counter()\n" ));
		
	/* Which counter? */
	minor = MINOR(inode->i_rdev);
	if ((counter = GETCOUNTER( inode->i_rdev )) == NULL)
		return -(EINVAL);

	return Open8253( counter );
}

STATIC void release_counter( struct inode *inode, struct file *file )
{
	Counter_t *counter;

	dprintk(( "LabPC+: release_counter()\n" ));
		
	if ((counter = GETCOUNTER( inode->i_rdev )) == NULL)
		return;

	Close8253( counter );
	labpc_close( inode, file );
}

struct file_operations counter_fops = {
	NULL,				/* seek */
	read_counter,
	write_counter,
	NULL,				/* readdir */
	NULL,				/* select_counter */
	ioctl_counter,
	NULL,				/* mmap */
	open_counter,
	release_counter,
};

#define LSBMSB		(RL0 | RL1)

int Write8253( Counter_t *counter, unsigned int data )
{
	unsigned char	c,	lo,	hi;

	/* Load the mode byte */
	c = ((LSBMSB & 0x30) |
	     ((counter->num << 6) & 0xc0) |
	     ((counter->mode << 1) & 0x0e) |
	     ((counter->bcd) & 0x01));
	counter->status |= COUNTER_COUNTING;
	
	/* Write the configuration to the mode port */
	outb_p( c, counter->mode_port );

	/* Write value to the counter - LSB then MSB */
	lo = data & 0xff;
	outb_p( lo, counter->data_port );
	hi = (data >> 8) & 0xff;
	outb_p( hi, counter->data_port );

	return 0;
}

unsigned int Read8253( Counter_t *counter )
{
	int	lsb,	msb;
	
	/* Load the mode port with a Counter Latch Command (RL0 = RL1 = 0) */
	outb_p( (counter->num << 6), counter->mode_port );

	/* Read value from the counter - LSB then MSB */
	lsb = inb_p( counter->data_port );
	msb = inb_p( counter->data_port );
	
	return (((msb << 8) & 0xff00) | (lsb & 0xff));
}

int SetLow8253( Counter_t *counter )
{
	unsigned char	c;
	
	/* Load the mode byte */
	c = ((LSBMSB & 0x30) |
	     ((counter->num << 6) & 0xc0) |
	     ((MODE_TERMINAL_COUNT << 1) & 0x0e) |
	     ((counter->bcd) & 0x01));

	outb_p( c, counter->mode_port );

	return 0;
}

int SetHigh8253( Counter_t *counter )
{
	unsigned char	c;
	
	/* Load the mode byte */
	c = ((LSBMSB & 0x30) |
	     ((counter->num << 6) & 0xc0) |
	     ((MODE_RATE_GENERATOR << 1) & 0x0e) |
	     ((counter->bcd) & 0x01));

	outb_p( c, counter->mode_port );

	return 0;
}

int SetMode8253( Counter_t *counter, int mode )
{
	/*	
	 * Set the mode for this counter. This will not affect
	 * currently running counts. The new mode will be implemented
	 * on the next "write" to the counter
	 */

	counter->mode = mode;

	return 0;			/* NoERR */
}

int SetBcd8253( Counter_t *counter, int bcd )
{
	counter->bcd = (bcd != 0);

	return 0;
}

int Open8253( Counter_t *counter )
{
	/* Initialise the counter structure */
	if (counter->first_time) {
		counter->first_time = 0;
		
		counter->status = 0;	/* Initialised			*/
		counter->mode = 2;	/* Set to mode 2 counting	*/
		counter->bcd = 0;	/* Binary mode counting		*/
		counter->open = 0;	/* We are open for business	*/
	}

	counter->open++;

	return 0;
}

int Close8253( Counter_t *counter )
{
	counter->status = 0;
	counter->open--;

	return 0;
}

int Init8253( Counter_t *counter, int minor, int num, int data_port, int mode_port )
{
	counter->first_time = 1;
	counter->data_port = data_port;
	counter->mode_port = mode_port;
	counter->num = num;
	counter->minor = minor;
	counter->mode = 2;
	counter->bcd = 0;
	counter->status = 0;
	counter->open = 0;

	return 0;
}
