/*
 *	$Id: labpcinit.c,v 0.5 1996/11/07 14:09:49 glenn Exp glenn $
 */

#include <linux/config.h>

#include <linux/module.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <stdlib.h>

#include "labpc.h"
#include "adc.h"
#include "dac.h"
#include "counter.h"
#include "dio.h"

#define MAXINTERRUPTLOOP	10

#ifndef LABPC_MAJOR
#define LABPC_MAJOR		21
#endif

#ifndef MAXDEVICES
#define MAXDEVICES		4
#endif

static char rcsrevision[] = "$Revision: 0.5 $";

/* Use these for initialising one LabPC card on the insmod cmd line */
int labpc_major = LABPC_MAJOR;
int labpc_base = LABPC_BASE;
int labpc_irq = LABPC_IRQ;
int labpc_dma = LABPC_DMA;

/* Use this for initialising more than one LabPC card on the insmod cmd line */
struct {
	int	major;
	int	base;
	int	irq;
	int	dma;
} labpc[MAXDEVICES] = {
	{ 0, 0, 0, 0 }
};

LabPC_t *boards[MAXDEVICES+1];

LabPC_t *getboard( int major )
{
	LabPC_t **board;

	for (board = boards; *board; board++) {
		if ((*board)->major == major) {
			return *board;
		}
	}

	dprintk(( "LabPC+: getboard unable to find major %d\n", major ));

	return NULL;
}

static int labpc_open( struct inode *inode, struct file *file )
{
	LabPC_t	*board = getboard( MAJOR( inode->i_rdev ) );
	int	minor = MINOR( inode->i_rdev );
	int	stat = 0;

	dprintk(( "LabPC+ open called with minor %d\n", minor ));

	if (board == NULL)
		return -(ENODEV);

	if (DAQ_COUNTERA0 <= minor && minor <= DAQ_COUNTERB2) {
		file->f_op = &counter_fops;
	} else if (minor == DAQ_DAC0 || minor == DAQ_DAC1) {
		file->f_op = &dac_fops;
	} else if (minor == DAQ_ADC) {
		file->f_op = &adc_fops;
	} else if (DAQ_DPORTA <= minor && minor <= DAQ_DPORTC) {
		file->f_op = &dio_fops;
	} else if (minor != DAQ_LABPC) {
		return -(ENODEV);
	}
	stat = file->f_op->open( inode, file );
	MOD_INC_USE_COUNT;

	return stat;
}

void labpc_close( struct inode *inode, struct file *file )
{
	MOD_DEC_USE_COUNT;
}

int labpc_ioctl_check( unsigned int cmd, unsigned long arg )
{
	int	len,	err;

	if (cmd & IOC_INOUT) {
		/* Have to validate the address given by the process. */
		len = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
		if ((cmd & IOC_IN) &&
		    ((err = verify_area (VERIFY_READ, (void *) arg, len)) < 0))
			return err;
		if ((cmd & IOC_OUT) &&
		    ((err = verify_area (VERIFY_WRITE, (void *) arg, len)) < 0))
			return err;
	}

	return 0;
}

static int labpc_ioctl( struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg )
{
	LabPC_t	*board = getboard( MAJOR( inode->i_rdev ) );
	int	stat = 0;
	
	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	switch (cmd) {
	case LABPC_STATUS:
		/* Enable/disable ADC input channel scanning */
		dprintregs( board );
		break;
	default:
		stat = -(ENOSYS);
		break;
	}

	return stat;
}
			
static struct file_operations labpc_fops = {
        NULL,		/* seek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	labpc_ioctl,	/* ioctl */
	NULL,		/* mmap */
        labpc_open,
        labpc_close	/* release */
};

/* The Interrupt Service Routine. Note that interrupt lines can be
 * shared between multiple LabPC cards */

static void daq_interrupt( int irq, void *dev_id, struct pt_regs *registerp )
{
	LabPC_t	*board = (LabPC_t *)dev_id;
	int	i,	status;

	i = 0;
	for (i = 0; i < MAXINTERRUPTLOOP; i++) {
		status = readport( &(board->regs.stat) );
		if (status & (ADC_OVERRUN | ADC_OVERFLOW | DMATC)) {
			adc_interrupt( board->adc, status );
		} else if ((status & DAVAIL) && !board->adc->dmaen) {
			adc_interrupt( board->adc, status );
		} else if (status & CNTINT) {
			dac_interrupt( board->dac, NUM_DACS );
		} else if ((board->regs.cmd3.value & DIOINTEN) &&
			   ((status = inb_p( board->dio->portC )) &
			    DIO_INTERRUPT)) {
			dio_interrupt( board->dio, status );
		} else {
			break;
		}
	}
	if (i >= MAXINTERRUPTLOOP) {
		printk( "LabPC+: ISR looped %d times: status = 0x%x - terminating\n",
			i, status );
	}
}

/* Initialise the Lab-PC+ Card (Manual 5-1). This is dodgy - we assume
 * card is present before testing if this is so. However, there
 * appears to be no reliable detection method, without first
 * initialising the card.
 */

int test_board( LabPC_t *board )
{
	CmdRegs_t	*regs = &(board->regs);
	int		stat;

	writeport( &(regs->cmd1), 0x00 ); /* ADC gain & input channel	*/
	writeport( &(regs->cmd2), 0x00 ); /* DAC config & ADC trigger	*/
	writeport( &(regs->cmd3), 0x00 ); /* Disable interrupts & DMA	*/
	writeport( &(regs->cmd4), 0x00 ); /* Disable ADC interval scanning*/
	outb_p( 0x34, board->base + DAQ_CTAMODE ); /* Counter A0 output = HI */
	outb_p( 0x0a, board->base + DAQ_CTA0 );	/* Load init A0 counter value */
	outb_p( 0x00, board->base + DAQ_CTA0 );
	outb_p( 0x74, board->base + DAQ_CTAMODE );
	outb_p( 0xb4, board->base + DAQ_CTAMODE );
	outb_p( 0x34, board->base + DAQ_CTBMODE );
	outb_p( 0x74, board->base + DAQ_CTBMODE );
	outb_p( 0xb4, board->base + DAQ_CTBMODE );
	outb_p( 0x00, board->base + DAQ_DMATCCLR );
	outb_p( 0x00, board->base + DAQ_TICLR );
	outb_p( 0x00, board->base + DAQ_ADCCLR );	/* ADC Clear	*/
	inb_p( board->base + DAQ_FIFO );
	inb_p( board->base + DAQ_FIFO );
	outb_p( 0x00, board->base + DAQ_DAC0LO );
	outb_p( 0x08, board->base + DAQ_DAC0HI );
	outb_p( 0x00, board->base + DAQ_DAC1LO );
	outb_p( 0x08, board->base + DAQ_DAC1HI );

	/* Check the Data available flag has been cleared */
	if ((stat = readport( &(regs->stat) )) & DAVAIL) {
		printk( "Lab-PC+: Failed attempt to initialise board at base address: 0x%x\n",
		        board->base );
		return -(EIO);
	}

	return 0;
}

int init_device( LabPC_t *board, int major, int base, int irq, int dma )
{
	int 		i;
	CmdRegs_t 	*regs;

	board->loaded  = 0;
	board->major   = major;
	board->base    = base;
	board->irq     = irq;
	board->dma     = dma;
	board->stat    = 0;
	board->open    = 0;
	board->present = 1;

	regs = &(board->regs);

	/* Check access to the io ports */
	if (check_region( board->base, LABPCIOSIZE )) {
		printk( "LabPC+: IO ports 0x%x to 0x%x already in use\n",
		        board->base, board->base + LABPCIOSIZE - 1 );
		printk( "LabPC+: Driver unloading\n" );

		return -(EIO);
	}

	/* Load the addresses of the command ports */
	regs->stat.address = board->base + DAQ_STAT;
	regs->cmd1.address = board->base + DAQ_CMD1;
	regs->cmd2.address = board->base + DAQ_CMD2;
	regs->cmd3.address = board->base + DAQ_CMD3;
	regs->cmd4.address = board->base + DAQ_CMD4;

	regs->stat.value = 0;

	if ((i = test_board( board )) != 0)
		return i;

	board->counter = (Counter_t *)kmalloc( sizeof( Counter_t )
					      * NUM_COUNTERS, GFP_KERNEL );
	if (board->counter == NULL) {
		kfree( board );
		return -(ENOMEM);
	}

	for (i = 0; i < 3; i++) {
		Init8253( &(board->counter[i]),
			 DAQ_COUNTERA0 + i,
			 i,
			 board->base + DAQ_CTA0 + i,
			 board->base + DAQ_CTAMODE );
		board->counter[i].board = board;
		Init8253( &(board->counter[3+i]),
			 DAQ_COUNTERB0 + i,
			 i,
			 board->base + DAQ_CTB0 + i,
			 board->base + DAQ_CTBMODE );
		board->counter[3+i].board = board;
	}

	board->dac = (Dac_t *)kmalloc( sizeof( Dac_t ) * NUM_DACS, GFP_KERNEL );

	if (board->dac == NULL) {
		kfree( board->counter );
		kfree( board );
		return -(ENOMEM);
	}

	for (i = 0; i < NUM_DACS; i++) {
		board->dac[i].first_time = 1;
		board->dac[i].board = board;
	}

	if ((board->adc = get_adc( board, DAQ_ADC )) == NULL) {
		kfree( board->counter );
		kfree( board->dac );
		kfree( board );
		return -(ENOMEM);
	}
	board->adc->board = board;

	board->dio = (Dio_t *)kmalloc( sizeof( Dio_t ), GFP_KERNEL );
	if (board->dio == NULL) {
		kfree( board->counter );
		kfree( board->dac );
		kfree( board->adc );
		kfree( board );
		return -(ENOMEM);
	}
	InitDio( board->dio, DAQ_DPORTA, board->base + DAQ_DIGA,
		 board->base + DAQ_DIGB, board->base + DAQ_DIGC,
		 board->base + DAQ_DIGMODE );
	board->dio->board = board;

	/* Reserve the ports for our use */
	request_region( board->base, LABPCIOSIZE, "labpc" );

	printk( "Lab-PC+ driver module %s\n", rcsrevision );
	if (register_chrdev( board->major, "labpc", &labpc_fops )) {
		release_region( board->base, LABPCIOSIZE );
		printk( "LabPC+: Unable to get major %d for Lab-PC+ devices\n",
		       board->major );
		return -(EINVAL);
	} else {
		printk( "Lab-PC+: Card detected at base address 0x%x\n", board->base );
	}
	if (request_irq( board->irq, daq_interrupt,
			 SA_INTERRUPT, "labpc", board )) {
		release_region( board->base, LABPCIOSIZE );
		unregister_chrdev( board->major, "labpc" );
		if (board->dma)
			free_dma( board->dma );
		printk( "LabPC+: can't request IRQ %d\n", board->irq );
		return -(EIO);
	}
	board->loaded = 1;
	
	return 0;
}

void release_device( LabPC_t *board )
{
	free_adc( board->adc );
	dprintk(( "Lab-PC+: ADC released\n" ));
	
	free_irq( board->irq, board );
	dprintk(( "Lab-PC+: irq released\n" ));

	release_region( board->base, LABPCIOSIZE );
	dprintk(( "Lab-PC+: ioports released\n" ));

	if (unregister_chrdev( board->major, "labpc" ) != 0) {
		printk( "Lab-PC+: device busy or other module error \n" );
	} else {
		printk( "Lab-PC+: succesfully removed \n" );
	} 
	dprintk(( "Lab-PC+: device major released\n" ));

	kfree( board->counter );
	dprintk(( "Lab-PC+: counter released\n" ));
	kfree( board->dac );
	dprintk(( "Lab-PC+: DAC released\n" ));
	kfree( board );
	printk( "Lab-PC+: Board released\n" );
}

void release_devices( LabPC_t **boards )
{
	LabPC_t **board;

	for (board = boards; *board != NULL; board++) {
		if ((*board)->loaded)
			release_device( *board );
	}
}

int init_module( void )
{
	int	i;
	int	stat;

	for (i = 0; i < MAXDEVICES+1; i++) {
		boards[i] = NULL;
	}

	/* If labpc array not initialised, load from defaults */
	if (labpc[0].major == 0) {
		labpc[0].major = labpc_major;
		labpc[0].base = labpc_base;
		labpc[0].irq = labpc_irq;
		labpc[0].dma = labpc_dma;
	}
	stat = -(EIO);
	for (i = 0; i < MAXDEVICES; i++) {
		if (labpc[i].major == 0)
			break;
		boards[i] = (LabPC_t *)kmalloc( sizeof( LabPC_t ), GFP_KERNEL );
		if (boards[i] == NULL)
			return -(ENOMEM);
		stat = init_device( boards[i],
				    labpc[i].major, labpc[i].base,
				    labpc[i].irq,   labpc[i].dma );
		if (stat != 0)
			release_devices( boards );
	}

	return stat;
}

void cleanup_module( void )
{
	if (MOD_IN_USE) {
		printk("Lab-PC+: device busy, remove delayed\n");
	}

	release_devices( boards );
}
