# Initialise the array to hold the config variables
proc initdacdata {dev} {
    set arr device$dev
    global ${arr}
    global config${arr}

    # Create the data array of configuration variables
    foreach i [$dev configure] {
	set ${arr}([string trimleft [lindex $i 0] "-"]) [lindex $i 1]
    }
    setdactimebase ${arr} {} w
    set config${arr}(data) {}

    trace variable ${arr} w "configdac $dev"
    trace variable ${arr}(clock) w "setdactimebase"
    trace variable ${arr}(timebase) w "setdactimebase"
}

# Trace the device config variables and perform required commands
proc configdac {device arr name op} {
    global $arr

    if {$op == "w"} {
	$device configure -$name [set ${arr}($name)]
    }
}

# Update the textvariable for displaying timebase & samplerate info
proc setdactimebase {arr name op} {
    global $arr
    global config$arr

    set timebase [set ${arr}(timebase)]
    set clock [set ${arr}(clock)]
    set period [expr $clock * $timebase / 2.0]

    # Choose the most appropriate units for the timebase display
    if {$period < 1} {
	set T_units {microseconds}
	set T_value $period
	set F_units {kHz}
	set F_value 0
    } elseif {$period < 2000} {
	set T_units {microseconds}
	set T_value $period
	set F_units {kHz}
	set F_value [expr 1000.0 / $period]
    } elseif {$period < 2000000} {
	set T_units {milliseconds}
	set T_value [expr $period / 1000.0]
	set F_units {Hz}
	set F_value [expr 1000000.0 / $period]
    } else {
	set T_units {seconds}
	set T_value [expr $period / 1000000.0]
	set F_units {Hz}
	set F_value [expr 1000000.0 / $period]
    }

    # Set the display message string
    set config${arr}(fstring) "Frequency: $F_value $F_units\nPeriod: $T_value $T_units"
}

# Make the menubar widget for the ADC
proc makedacmenubar {w arr} {
    global ${arr}

    # Build the DAC menubar
    frame $w -borderwidth 2 -relief raised

    # The File menu
    set m $w.filemenu
    menubutton $m \
	    -menu "$m.m" \
	    -text {File}
    menu $m.m
    $m.m add command \
	    -command "destroy [winfo toplevel $w]" \
	    -label {Quit}
 
    # The Help Menu
    set m $w.helpmenu
    menubutton $m \
	    -menu "$m.m" \
	    -text {Help}
    menu $m.m
    $m.m add command \
	    -label {LabPC} \
	    -command "popuphelp labpc"
    $m.m add command \
	    -label {DAC} \
	    -command "popuphelp DAC"

    # Pack them all together
    pack append $w \
	    $w.filemenu {left frame center} \
	    $w.helpmenu {right frame center}
}

proc DacWrite {dac} {
    set arr device${dac}
    global ${arr}
    global config${arr}

    $dac write [set config${arr}(data)]
}

proc Plotfunction {graph arr txt} {
    global ${arr}
    global config${arr}

    set clock [set ${arr}(clock)]
    set timebase [set ${arr}(timebase)]
    set nsamples [set config${arr}(nsamples)]
    set period [expr $clock * $timebase / 2.0]
    if {$nsamples*$period < 2000} {
	set dt $period
	$graph configure -xlabel "Time (microseconds)"
    } elseif {$nsamples*$clock < 2000000} {
	set dt [expr $period / 1000.0]
	$graph configure -xlabel "Time (milliseconds)"
    } else {
	set dt [expr $period / 1000000.0]
	$graph configure -xlabel "Time (seconds)"
    }
    set period [expr $period / 1000000.0]
    set config${arr}(data) {}
    set xdata {}
    for {set i 0} {$i < $nsamples} {incr i} {
	set t [expr $i * $period]
	set f [expr double($i) / $nsamples]
	lappend config${arr}(data) [expr $txt]
	lappend xdata [expr $i * $dt]
    }

    # Delete any old displayed graphs
    foreach i [$graph names] {
	$graph delete $i
    }

    $graph insert dac0 \
	    -xdata $xdata \
	    -ydata [set config${arr}(data)] \
	    -label "DAC output"
}

proc makedacdisplay {w dev arr} {
    global ${arr}
    global config${arr}

    # Build the analogue display widget
    frame $w

    # Build the graph widget (a frame wrapper so we can set the class)
    frame $w.graph \
	    -class Graph
    xygraph $w.graph.display
    set bg [option get $w.graph.display background Background]
    if {"$bg" == ""} {
	$w.graph.display configure -background [$w cget -background]
    } else {
	$w.graph.display configure -background "$bg"
    }
    pack append $w.graph \
	    $w.graph.display {left frame center fill expand}

    # Build the control panel
    set p $w.panel
    frame $p \
	    -relief raised \
	    -borderwidth 2

    # DAC write button
    button $p.write \
	    -text "Write" \
	    -command "DacWrite $dev"
    bind $p.write <Any-ButtonPress-3> "popuphelp \"dac write\""

    # Number of Samples
    set config${arr}(nsamples) 100
    labelentry $p.nsamples "Number of Samples" config${arr}(nsamples)
    set config${arr}(nsamples) [option get $p.nsamples nsamples Nsamples]

    # Polarity
    combo $p.polarity ${arr} ${arr}(polarity) "Polarity" {bipolar unipolar}

    # Mode
    combo $p.mode ${arr} ${arr}(mode) "Mode" {wave trace write}

    # Text widget for function defn
    frame $p.func
    text $p.func.text \
	    -setgrid yes \
	    -yscrollcommand "$p.func.scroll set"
    scrollbar $p.func.scroll \
	    -command "$p.func.text yview"
    button $p.button \
	    -text {Apply} \
	    -command "Plotfunction $w.graph.display $arr \"\[$p.func.text get 0.0 end\]\""

    pack append $p.func \
	    $p.func.scroll {right filly} \
	    $p.func.text {left expand fill}

    # Pack the buttons into the control panel
    pack append $p \
	    $p.write		{top pady 5} \
	    $p.nsamples		{top pady 5} \
	    $p.polarity		{top fillx pady 5} \
	    $p.mode		{top fillx pady 5} \
	    $p.func		{top expand fill} \
	    $p.button		{top}


    # Pack the display widget
    pack append $w \
	    $w.graph {left frame center fillx filly expand} \
	    $w.panel {right frame center filly}
}

# Create a widget for displaying the timebase information
proc makedactimebase {w arr} {
    global ${arr}
    global config${arr}

    # Build the control panel
    frame $w \
	    -relief raised \
	    -borderwidth 2

    labelentry $w.clock "DAC Clock\n(counts)" "${arr}(clock)"
    
    labelentry $w.timebase "Clock Timebase\n(microseconds)" "${arr}(timebase)"

    label $w.summary \
	    -justify left \
	    -textvariable "config${arr}(fstring)"
    
    pack append $w \
	    $w.clock {left} \
	    $w.timebase {left} \
	    $w.summary {left}
}

# procedure to show window
proc dacwidget {w dac} {
    global device$dac
    set arr device$dac

    initdacdata $dac

    catch "destroy $w"

    # The encasing frame for the entire widget
    frame $w \
	    -class DAC

    # Build the widget menubar
    makedacmenubar $w.menubar $arr

    # Build the analogue display widget
    makedacdisplay $w.display $dac $arr

    # Make the timbase information display panel
    makedactimebase $w.timebase $arr

    pack append $w \
	    $w.menubar {top frame center fillx} \
	    $w.display {top fillx filly expand} \
	    $w.timebase {bottom frame center fillx}
}
