#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>

#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include <tcl.h>
#include <tk.h>
#include <labpc.h>

#ifndef DEVICE_DIR
#define	DEVICE_DIR	"/dev/labpc"
#endif

#define	IOCTL	1

#define MAX_ENUM_VALUES	10

typedef struct {
	char	*name;
	enum {
		Integer,
		Boolean,
		Enumerated,
	} type;
	int	flags;
	int	setioctl;
	int	getioctl;
	int	arg;
	struct {
		char	*name;
		int	value;
	} values[MAX_ENUM_VALUES];
} Option_t;

typedef struct {
	char	*name;
	char	*filename;
	int	fd;
	char	*handler;
	Tcl_Interp *handler_interp;
	Option_t *options;
} Device_t;

static char *device_dir = DEVICE_DIR;

static Option_t AdcOptions[] = {
	{ "raw",	Boolean, 0,	0,	0,	0, },
	{ "nonblock",	Boolean, 0,	0,	0,	0, },
	{ "timebaseunits",	Enumerated, 0,	0,	0,	1,
	  { {"microseconds", 	1	},
	    {"milliseconds", 	1000 	},
	    {"seconds",		1000000	}, }, },
	{ "autoacquire",Boolean, IOCTL, ADC_SET_AUTOACQUIRE,  ADC_GET_AUTOACQUIRE, 0, },
	{ "channel",	Integer, IOCTL, ADC_SET_CHANNEL,  ADC_GET_CHANNEL, 0, },
	{ "scan",	Boolean, IOCTL, ADC_SET_SCAN,	  ADC_GET_SCAN,	 0,},
	{ "dma",	Boolean, IOCTL, ADC_SET_DMA,	  ADC_GET_DMA,	 0,},
	{ "buffers",	Integer, IOCTL, ADC_SET_NBUF,     ADC_GET_NBUF, 0, },
	{ "buffersize", Integer, IOCTL, ADC_SET_BUFSIZE,  ADC_GET_BUFSIZE, 0, },
	{ "timebase",	Integer, IOCTL, ADC_SET_TIMEBASE, ADC_GET_TIMEBASE, 0, },
	{ "clock",	Integer, IOCTL, ADC_SET_CLOCK, 	  ADC_GET_CLOCK,   0,},
	{ "intscan",	Boolean, IOCTL, ADC_SET_INTSCAN,  ADC_GET_INTSCAN, 0, },
	{ "intclock",	Integer, IOCTL, ADC_SET_INTCLOCK, ADC_GET_INTCLOCK, 0, },
	{ "intcount",	Integer, IOCTL, ADC_SET_INTCOUNT, ADC_GET_INTCOUNT, 0, },
	{ "gain",	Enumerated, IOCTL, ADC_SET_GAIN,  ADC_GET_GAIN,	 0,
	  { {"1", 	ADC_GAIN_1	},
	    {"1.25", 	ADC_GAIN_1p25 	},
	    {"2",	ADC_GAIN_2	},
	    {"5", 	ADC_GAIN_5 	},
	    {"10", 	ADC_GAIN_10	},
	    {"20", 	ADC_GAIN_20	},
	    {"50", 	ADC_GAIN_50	},
	    {"100", 	ADC_GAIN_100	} } },
	{ "polarity",	Enumerated, IOCTL, ADC_SET_POLARITY, ADC_GET_POLARITY, 0,
	  { {"unipolar",ADC_UNIPOLAR	},
	    {"bipolar", ADC_BIPOLAR	} } },
	{ "trigger",	Enumerated, IOCTL, ADC_SET_TRIGGER_MODE, ADC_GET_TRIGGER_MODE, 0, 
	  { {"auto", 	ADC_AUTOTRIGGER	},
	    {"soft", 	ADC_SOFTTRIGGER	},
	    {"hard", 	ADC_HARDTRIGGER	},
	    {"pre", 	ADC_PRETRIGGER	} } },
	{ "connection",	Enumerated, IOCTL, ADC_SET_CONNECTION, ADC_GET_CONNECTION,  0,
	  { {"singleended", ADC_SINGLEENDED},
	    {"differential", ADC_DIFFERENTIAL} } },
	{ NULL,	}
};

static Option_t DacOptions[] = {
	{ "raw",	Boolean, 0,	0,		0,		 0, },
	{ "timebase",	Integer, IOCTL, DAC_SET_TIMEBASE, DAC_GET_TIMEBASE,  0,},
	{ "clock",	Integer, IOCTL, DAC_SET_CLOCK,	  DAC_GET_CLOCK,    0, },
	{ "polarity",	Enumerated, IOCTL,DAC_SET_POLARITY, DAC_GET_POLARITY,  0,
	  { {"unipolar",DAC_UNIPOLAR	},
	    {"bipolar", DAC_BIPOLAR	} } },
	{ "mode",	Enumerated, IOCTL, DAC_SET_MODE, DAC_GET_MODE,	  0,   
	  { {"wave",	DAC_WAVEFORM_MODE},
	    {"trace",	DAC_TRACE_MODE	},
	    {"write",	DAC_WRITE_MODE	} } },
	{ NULL,	}
};

static Option_t CtrOptions[] = {
	{ "mode",	   Enumerated, IOCTL, COUNTER_SET_MODE,  COUNTER_GET_MODE, 0,
	  { {"terminal",   MODE_TERMINAL_COUNT	},
	    {"oneshot",    MODE_ONE_SHOT	},
	    {"rate",	   MODE_RATE_GENERATOR	},
	    {"squarewave", MODE_SQUARE_WAVE	},
	    {"softstrobe", MODE_SOFT_STROBE	},
	    {"hardstrobe", MODE_HARD_STROBE	} } },
	{ NULL,	}
};

static Option_t DioOptions[] = {
	{ NULL, }
};

static char *ListOptions( Option_t *option )
{
	static char	sbuf[200];

	sbuf[0] = '\0';
	while (option->name) {
		strcat( strcat( strcat( sbuf, "-" ), option->name ), " " );
		option++;
	}

	return sbuf;
}

static Option_t *GetOption( Tcl_Interp *interp, Option_t *list, char *name )
{
	Option_t *option;

	for (option = list; option->name; option++) {
		if (!strcmp( option->name, name ))
			return option;
	}
	Tcl_AppendResult( interp,
			  "Unknown option ",
			  name,
			  " - should be one of:\n",
			  ListOptions( list ),
			  NULL );

	return NULL;
}

static char *ListArgValues( Tcl_Interp *interp, Option_t *option )
{
	static char	sbuf[200];
	int		i;

	switch (option->type) {
	case Integer:
		strcpy( sbuf, "number" );
		break;
	case Boolean:
		strcpy( sbuf, "on off yes no 1 0" );
		break;
	case Enumerated:
		sbuf[0] = '\0';
		for (i = 0; option->values[i].name != NULL; i++) {
			strcat( strcat( sbuf, option->values[i].name ), " ");
		}
		break;
	default:
		Tcl_SetResult( interp, "Labpc internal error: ListArgValues()",
			       TCL_STATIC );
		return NULL;
		break;
	}

	return sbuf;
}

static int GetArgValue( Tcl_Interp *interp, Option_t *option, char *s )
{
	int	arg;
	int	i;

	if (!s) {
		Tcl_AppendResult( interp,
				  option->name,
				  " not set: should be \"-",
				  option->name,
				  " [",
				  ListArgValues( interp, option ),
				  "]\"",
				  NULL );
		return TCL_ERROR;
	}
	
	switch (option->type) {
	case Integer:
		if (Tcl_GetInt( interp, s, &arg ) == TCL_OK) {
			return arg;
		}
		break;
	case Boolean:
		if (Tcl_GetBoolean( interp, s, &arg ) == TCL_OK) {
			return arg;
		}
		break;
	case Enumerated:
		for (i = 0; option->values[i].name; i++) {
			if (!strcmp( option->values[i].name, s ))
				return option->values[i].value;
		}
		break;
	default:
		Tcl_SetResult( interp, "Labpc internal error: GetArgValue()",
			       TCL_STATIC );
		return -1;
		break;
	}
		
	Tcl_AppendResult( interp,
			  option->name,
			  " bad value \"",
			  s, "\": should be one of ",
			  " [",
			  ListArgValues( interp, option ),
			  "]\"",
			  NULL );
	return -1;
}

static char *GetArgString( Tcl_Interp *interp, Option_t *option, int arg )
{
	static char	sbuf[200];
	int		i;

	switch (option->type) {
	case Integer:
		sprintf( sbuf, "%d", arg );
		break;
	case Boolean:
		if (arg) {
			strcpy( sbuf, "on" );
		} else {
			strcpy( sbuf, "off" );
		}
		break;
	case Enumerated:
		for (i = 0; option->values[i].name; i++) {
			if (option->values[i].value == arg) {
				strcpy( sbuf, option->values[i].name );
				break;
			}
		}
		if (!option->values[i].name) {
			Tcl_SetResult( interp,
				       "Labpc internal error:"
				       " bad value in GetArgString()",
				       TCL_STATIC );
			return NULL;
		}
		break;
	default:
		Tcl_SetResult( interp,
			       "Labpc internal error: in GetArgString()",
			       TCL_STATIC );
		return NULL;
		break;
	}

	return sbuf;
}

static int SetOptionValue( Tcl_Interp *interp, Device_t *device, Option_t *option, char *name, char *s )
{
	int	arg;

	if ((option = GetOption( interp, option, name )) == NULL)
		return 0;

	if ((arg = GetArgValue( interp, option, s )) < 0)
		return -1;

	if (option->flags & IOCTL) {
		if (ioctl( device->fd, option->setioctl, arg ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(set_", option->name, "): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return -1;
		}
	}

	option->arg = arg;

	return 1;
}

static int GetOptionValue(
			  Tcl_Interp *interp,
			  Device_t   *device,
			  Option_t   *option,
			  char 	     *name
			  )
{
	int	arg;

	if ((option = GetOption( interp, option, name )) == NULL)
		return -1;

	if (option->flags & IOCTL) {
		if (ioctl( device->fd, option->getioctl, &arg ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(get_", option->name, "): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return -1;
		}
	} else {
		arg = option->arg;
	}

	return arg;
}

static char *GetOptionString(
			     Tcl_Interp *interp,
			     Device_t   *device,
			     Option_t   *option,
			     char       *name
			     )
{
	int	arg;
	char	*s;

	if ((option = GetOption( interp, option, name )) == NULL)
		return NULL;
	if ((arg = GetOptionValue( interp, device, option, name )) < 0)
		return NULL;
	if ((s = GetArgString( interp, option, arg )) == NULL)
		return NULL;

	return s;
}

static int ConfigureAllOptions(
			       Tcl_Interp *interp,
			       Device_t   *device,
			       Option_t   *options
			       )
{
	Option_t *option;
	char	 sbuf[200];
	char	 *s;
	
	for (option = options; option->name != NULL; option++) {
		if ((s = GetOptionString( interp, device, options, option->name )) == NULL)
			return TCL_ERROR;
		sprintf( sbuf, "-%s %s", option->name, s );
		Tcl_AppendElement( interp, sbuf );
	}

	return TCL_OK;
}

static int ConfigureOption( Tcl_Interp *interp, Device_t *device, Option_t *options, char *name )
{
	char	sbuf[200];
	char	*s;

	if (name[0] != '-') {
		Tcl_AppendResult( interp,
				 "Unknown option ",
				 name,
				 " - should be one of:\n",
				 ListOptions( options ),
				 NULL );
		return TCL_ERROR;
	}
	name++;
	if ((s = GetOptionString( interp, device, options, name )) == NULL)
		return TCL_ERROR;
	
	sprintf( sbuf, "-%s %s", name, s );
	Tcl_SetResult( interp, sbuf, TCL_VOLATILE );

	return TCL_OK;
}

static int CGetOption( Tcl_Interp *interp, Device_t *device, Option_t *options, char *name )
{
	char	*s;

	if (name[0] != '-') {
		Tcl_AppendResult( interp,
				 "Unknown option ",
				 name,
				 " - should be one of:\n",
				 ListOptions( options ),
				 NULL );
		return TCL_ERROR;
	}
	name++;
	if ((s = GetOptionString( interp, device, options, name )) == NULL)
		return TCL_ERROR;

	Tcl_SetResult( interp, s, TCL_VOLATILE );

	return TCL_OK;
}

static int ProcessOptions(
			  Tcl_Interp 	*interp,
			  Device_t	*device,
			  int 		argc,
			  char 		**argv
			  )
{
	char	*name;

	/* Process configuration options */
	while (*argv) {
		name = *argv;
		if (name[0] != '-') {
			Tcl_AppendResult( interp,
					  "Unknown option ",
					  *argv,
					  " - should be one of:\n",
					  ListOptions( device->options ),
					  NULL );
			return TCL_ERROR;
		}
		++name;
		if (SetOptionValue( interp, device, device->options, name, *++argv ) < 0) {
			return TCL_ERROR;
		}
		++argv;
	}

	return TCL_OK;
}

static int ProcessCommand(
			  Tcl_Interp 	*interp,
			  Device_t	*device,
			  int 		argc,
			  char 		**argv
			  )
{
	/* Process commands common to all LabPC device cmds */
	if (!strcmp( *argv, "destroy" )) {
		/* Delete this command from the interpreter */
		if (Tcl_DeleteCommand( interp, device->name ) != TCL_OK) {
			return TCL_ERROR;
		}
		return TCL_OK;
	} else if (!strcmp( *argv, "configure" )) {
		/* Set configuration options - no readings */
		if (argc == 2) {
			/* Get a list of all option values */
			return ConfigureAllOptions( interp, device, device->options );
		} else if (argc == 3) {
			/* Get the named option value */
			return ConfigureOption( interp, device, device->options, argv[1] );
		} else {
			return ProcessOptions( interp, device, argc-1, argv+1 );
		}
	} else if (!strcmp( *argv, "cget" )) {
		/* Return a configuration option */
		if (argc != 3) {
			Tcl_AppendResult( interp,
					  "cget option not specified:",
					  "should be one of:",
					  ListOptions( device->options ),
					  NULL );
			return TCL_ERROR;
		}
		return CGetOption( interp, device, device->options, argv[1] );
	}

	return ProcessOptions( interp, device, argc, argv );
}

static void FileHandler( ClientData clientdata, int mask )
{
	Device_t *device = (Device_t *)clientdata;

	if (device->handler)
		Tcl_GlobalEval( device->handler_interp, device->handler );
}

static int AdcCmd( 
		  ClientData clientdata,
		  Tcl_Interp *interp,
		  int argc,
		  char **argv
		  )
{
	Device_t	*device = (Device_t *)clientdata;
	Tcl_DString	dspec;

	if (*++argv == NULL) {
		Tcl_SetResult( interp, "No cmd - should be one of: read, trigger, acquire, stop, init, handler, readtimebase, configure, cget, destroy", TCL_STATIC );
		return TCL_ERROR;
	}

	if (!strcmp( *argv, "read" )) {
		char		sbuf[100];
		int		num = 0;
		short		*dbuf;
		int		i,	n;
		double		cal = 1.0;
		int		raw;
		int		scan;
		int		nchan;
		int		nonblock;

		/* Get the number of readings to take */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "# of readings not specified:",
					  "should be \"",
					  "read number_of_readings\"",
					  NULL );
			return TCL_ERROR;
		}
		if (Tcl_GetInt( interp, argv[1], &num ) != TCL_OK) {
			return TCL_ERROR;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		/* Get the calibration factor for conversions */
		raw = GetOptionValue( interp, device,
				      device->options, "raw" );
		nonblock = GetOptionValue( interp, device,
					   device->options, "nonblock" );
		scan = GetOptionValue( interp, device,
				       device->options, "scan" );
		if (!raw) {
			i = GetOptionValue( interp, device,
					    device->options, "gain" );
			switch (i) {
			case ADC_GAIN_1:
				cal = 1.0;
				break;
			case ADC_GAIN_1p25:
				cal = 1.25;
				break;
			case ADC_GAIN_2:
				cal = 2.0;
				break;
			case ADC_GAIN_5:
				cal = 5.0;
				break;
			case ADC_GAIN_10:
				cal =10.0;
				break;
			case ADC_GAIN_20:
				cal = 20.0;
				break;
			case ADC_GAIN_50:
				cal = 50.0;
				break;
			case ADC_GAIN_100:
				cal = 100.0;
				break;
			default:
				return TCL_ERROR;
				break;
			}
			cal = (10.0 / 4096) / cal;
		}

		nchan = 1;
		if (scan) {
			nchan = GetOptionValue( interp, device, device->options, "channel" ) + 1;
			num *= nchan;
		}

		/* Allocate buffer for the array to read from ADC */
		dbuf = (short *)malloc( num * sizeof( short ) );
		if (dbuf == NULL) {
			Tcl_SetResult( interp, "Out of memory", TCL_STATIC );
			return TCL_ERROR;
		}

		if (nonblock) {
			if (fcntl( device->fd, F_SETFL,
				   fcntl( device->fd, F_GETFL ) | O_NONBLOCK ) < 0) {
				Tcl_AppendResult( interp, "fcntl()",
						  Tcl_PosixError( interp ),
						  NULL );
				free( dbuf );
				return TCL_ERROR;
			}
		} else {
			if (fcntl( device->fd, F_SETFL,
				   fcntl( device->fd, F_GETFL ) & ~O_NONBLOCK ) < 0) {
				Tcl_AppendResult( interp, "fcntl()",
						  Tcl_PosixError( interp ),
						  NULL );
				free( dbuf );
				return TCL_ERROR;
			}
		}

		/* Read the data from the ADC */
		num = read( device->fd, dbuf, num * sizeof( short ) );
		if (num < 0) {
			if ((errno == EWOULDBLOCK) || (errno == EAGAIN)) {
				num = 0;
			} else {
				Tcl_AppendResult( interp, "read(): ",
						  Tcl_PosixError( interp ),
						  NULL );
				free( dbuf );
				return TCL_ERROR;
			}
		}
		num /= sizeof( short );

		/* Convert raw data to string for Tcl */
		for (n = nchan - 1; n >= 0; n--) {
			Tcl_DStringInit(&dspec);
			for (i = n; i < num; i += nchan) {
				sprintf( sbuf, "%g", dbuf[i] * cal );
				Tcl_DStringAppendElement( &dspec, sbuf );
			}
			Tcl_AppendElement( interp, dspec.string );
			Tcl_DStringFree( &dspec );
		}
		free( dbuf );
	} else if (!strcmp( *argv, "readtimebase" )) {
		unsigned long period,	interval;
		char	sbuf[100];
		int	i,	j,	n;
		int	num = 0;
		int	clock,		timebase,
			intscan,	intclock,
			intcount,	nchan,
			scan;
		double	units;

			
		/* Get the number of readings to take */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "# of readings not specified:",
					  "should be \"",
					  "readtimebase number_of_readings\"",
					  NULL );
			return TCL_ERROR;
		}
		if (Tcl_GetInt( interp, argv[1], &num ) != TCL_OK) {
			return TCL_ERROR;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		/* Get the various option settings */
		clock = GetOptionValue( interp, device,
					device->options, "clock" );
		timebase = GetOptionValue( interp, device,
					   device->options, "timebase" );
		intscan = GetOptionValue( interp, device,
					  device->options, "intscan" );
		intclock = GetOptionValue( interp, device,
					   device->options, "intclock" );
		intcount = GetOptionValue( interp, device,
					   device->options, "intcount" );
		units = GetOptionValue( interp, device,
					device->options, "timebaseunits" );
		scan = GetOptionValue( interp, device,
				       device->options, "scan" );
		nchan = 1;
		if (scan) {
			nchan = GetOptionValue( interp, device, device->options, "channel" ) + 1;
			num *= nchan;
		}

		/* Check we can determine the timebase ie. no external timing */
		if (clock == 0) {
			Tcl_SetResult( interp, "External clock source selected - no timebase information",
				       TCL_STATIC );
			return TCL_ERROR;
		}
		if (intscan && intclock == 0) {
			Tcl_SetResult( interp, "External interval clock source selected - no timebase information",
				       TCL_STATIC );
			return TCL_ERROR;
		}

		if (intscan) {
			period = (clock * timebase) / 2.0;
			interval = (intclock * timebase) / 2.0;
			if (nchan > 1) {
				for (n = 0; n < nchan; n++) {
					Tcl_DStringInit(&dspec);
					j = 0;
					for (i = nchan - n - 1; i < num; i += nchan) {
						sprintf( sbuf, "%g", (i * period + j++ * interval) / units );
						Tcl_DStringAppendElement( &dspec, sbuf );
					}
					Tcl_AppendElement( interp, dspec.string );
					Tcl_DStringFree( &dspec );
				}
			} else {
				Tcl_DStringInit(&dspec);
				i = j = 0;
				for (n = 0; n < num; n++) {
					sprintf( sbuf, "%g",
						 (i * interval + j * period) / units );
					Tcl_DStringAppendElement( &dspec, sbuf );
					if (++j % intcount == 0) {
						j = 0;
						i++;
					}
				}
				Tcl_AppendElement( interp, dspec.string );
				Tcl_DStringFree( &dspec );
			}
		} else {
			period = (clock * timebase) / 2.0;
			for (n = 0; n < nchan; n++) {
				Tcl_DStringInit(&dspec);
				for (i = nchan - n - 1; i < num; i += nchan) {
					sprintf( sbuf, "%g", (i * period) / units );
					Tcl_DStringAppendElement( &dspec, sbuf );
				}
				Tcl_AppendElement( interp, dspec.string );
				Tcl_DStringFree( &dspec );
			}
		}
	} else if (!strcmp( *argv, "acquire" )) {
		int	scan,	nchan,	num = 0;

		/* Get the number of readings to take */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "# of readings not specified:",
					  "should be \"",
					  "acquire number_of_readings\"",
					  NULL );
			return TCL_ERROR;
		}
		if (Tcl_GetInt( interp, argv[1], &num ) != TCL_OK) {
			return TCL_ERROR;
		}
		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;


		scan = GetOptionValue( interp, device,
				       device->options, "scan" );
		nchan = 1;
		if (scan) {
			nchan = GetOptionValue( interp, device, device->options, "channel" ) + 1;
			num *= nchan;
		}

		if (ioctl( device->fd, ADC_ACQUIRE, num ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(ADC_ACQUIRE): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return TCL_ERROR;
		}
	} else if (!strcmp( *argv, "handler" )) {
		if (!argv[1]) {
			if (device->handler) {
				Tcl_SetResult( interp, device->handler,
					       TCL_VOLATILE );
			} else {
				Tcl_SetResult( interp, "", TCL_STATIC );
			}
			return TCL_OK;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;
		if (device->handler) {
			Tk_DeleteFileHandler( (ClientData) device->fd );
			device->handler = NULL;
		}
		if (strlen( argv[1] ) > 0) {
			device->handler = strdup( argv[1] );
			device->handler_interp = interp;
			Tk_CreateFileHandler( (ClientData) device->fd,
					      TK_READABLE | TK_EXCEPTION,
					      FileHandler,
					      (ClientData) device );
		}
	} else if (!strcmp( *argv, "trigger" )) {
		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-1, argv+1 ) != TCL_OK)
			return TCL_ERROR;

		if (ioctl( device->fd, ADC_TRIGGER ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(ADC_TRIGGER): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return TCL_ERROR;
		}
	} else if (!strcmp( *argv, "stop" )) {
		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-1, argv+1 ) != TCL_OK)
			return TCL_ERROR;

		if (ioctl( device->fd, ADC_STOP ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(ADC_STOP): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return TCL_ERROR;
		}
	} else if (!strcmp( *argv, "status" )) {
		unsigned long	status;

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-1, argv+1 ) != TCL_OK)
			return TCL_ERROR;

		if (ioctl( device->fd, ADC_GET_STATUS, &status ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(ADC_STATUS): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return TCL_ERROR;
		}
		if (status & ADC_STATUS_ACTIVE)
			Tcl_AppendElement( interp, "active" );
		if (status & ADC_STATUS_INACTIVE)
			Tcl_AppendElement( interp, "inactive" );
		if (status & ADC_STATUS_STOPPED)
			Tcl_AppendElement( interp, "stopped" );
		if (status & ADC_STATUS_STARTING)
			Tcl_AppendElement( interp, "starting" );
		if (status & ADC_STATUS_OVERRUN)
			Tcl_AppendElement( interp, "overrun" );
		if (status & ADC_STATUS_OVERFLOW)
			Tcl_AppendElement( interp, "overflow" );
		if (status & ADC_STATUS_BUFFER_OVERFLOW)
			Tcl_AppendElement( interp, "buffer_overflow" );
		if (status & ADC_STATUS_IO_ERROR)
			Tcl_AppendElement( interp, "io_error" );
		if (status & ADC_STATUS_ACQUISITION_COMPLETE)
			Tcl_AppendElement( interp, "acquisition_complete" );
		if (status & ADC_STATUS_ERROR)
			Tcl_AppendElement( interp, "error" );
	} else if (!strcmp( *argv, "init" )) {
		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-1, argv+1 ) != TCL_OK)
			return TCL_ERROR;

		if (ioctl( device->fd, ADC_INIT ) < 0) {
			Tcl_AppendResult( interp,
					 "ioctl(ADC_INIT): ",
					 Tcl_PosixError( interp ),
					 NULL );
			return TCL_ERROR;
		}
	} else {
		return ProcessCommand( interp, device, argc, argv );
	}

	return TCL_OK;
}

static int DacCmd( 
		  ClientData clientdata,
		  Tcl_Interp *interp,
		  int argc,
		  char **argv
		  )
{
	Device_t	*device = (Device_t *)clientdata;
	int		num = 0;
	short		*dbuf;
	int		i;
	int		listc;
	char		**listv;
	double		d;
	int		raw;
	int		polarity;
	int		lowlimit,	highlimit;
	char		sbuf[100];

	if (*++argv == NULL) {
		Tcl_SetResult( interp, "No cmd - should be one of: read, write, configure, cget", TCL_STATIC );
		return TCL_ERROR;
	}

	if (!strcmp( *argv, "write" )) {
		/* Get list of values to write */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "Nothing to write: should be "
					  "\"write {n1 ?n2 n3 ...?}\"",
					  NULL );
			return TCL_ERROR;
		}
		if (Tcl_SplitList( interp, argv[1], &listc, &listv ) != TCL_OK) {
			return TCL_ERROR;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		raw = GetOptionValue( interp, device, device->options, "raw" );
		polarity = GetOptionValue( interp, device,
					   device->options, "polarity" );
		if (polarity == DAC_UNIPOLAR) {
			lowlimit = 0;
			highlimit = 4095;
		} else {
			lowlimit = -2048;
			highlimit = 2047;
		}

		/* Allocate buffer for data to write to DAC */
		dbuf = (short *)malloc( listc * sizeof( short ) );
		if (dbuf == NULL) {
			Tcl_SetResult( interp, "Out of memory", TCL_STATIC );
			return TCL_ERROR;
		}

		/* Convert the Tcl strings to raw data */
		for (i = 0; i < listc; i++) {
			if (Tcl_GetDouble( interp, listv[i], &d ) != TCL_OK) {
				return TCL_ERROR;
			}
			if (!raw)
				d *= (4096.0/10.0);
			if (d < 0.0) {
				if ((dbuf[i] = (short)(d - 0.5)) < lowlimit)
					dbuf[i] = lowlimit;
			} else {
				if ((dbuf[i] = (short)(d + 0.5)) > highlimit)
					dbuf[i] = highlimit;
			}
		}

		/* Write the data to the device */
		if ((num = write( device->fd, dbuf,
				  listc * sizeof( short ) )) < 0) {
			Tcl_AppendResult( interp, "write()",
					  Tcl_PosixError( interp ),
					  NULL );
			free( dbuf );
			free( listv );
			return TCL_ERROR;
		}
		free( dbuf );
		free( listv );

		sprintf( sbuf, "%d", num );
		Tcl_SetResult( interp, sbuf, TCL_VOLATILE );
	} else {
		return ProcessCommand( interp, device, argc, argv );
	}

	return TCL_OK;
}

static int CtrCmd( 
		  ClientData clientdata,
		  Tcl_Interp *interp,
		  int argc,
		  char **argv
		  )
{
	Device_t	*device = (Device_t *)clientdata;
	short		val;

	if (*++argv == NULL) {
		Tcl_SetResult( interp, "No cmd - should be one of: read, write, configure, cget or destroy", TCL_STATIC );
		return TCL_ERROR;
	}

	if (!strcmp( *argv, "read" )) {
		char	sbuf[200];

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-1, argv+1 ) != TCL_OK)
			return TCL_ERROR;

		/* Read the value from the device */
		if (read( device->fd, &val, sizeof( short ) ) < 0) {
			Tcl_AppendResult( interp, "read()",
					  Tcl_PosixError( interp ),
					  NULL );
			return TCL_ERROR;
		}
		sprintf( sbuf, "%d", (int)val );

		Tcl_SetResult( interp, sbuf, TCL_VOLATILE );
	} else if (!strcmp( *argv, "write" )) {
		int	num;
		
		/* Get the value to write to the counter */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "Counter setting not specified:"
					  "should be \"write number\"",
					  NULL );
			return TCL_ERROR;
		}
		if (Tcl_GetInt( interp, argv[1], &num ) != TCL_OK) {
			return TCL_ERROR;
		}
		val = (short)num;

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		/* Write the data to the device */
		if (write( device->fd, &val, sizeof( short ) ) < 0) {
			Tcl_AppendResult( interp, "write()",
					  Tcl_PosixError( interp ),
					  NULL );
			return TCL_ERROR;
		}
	} else {
		return ProcessCommand( interp, device, argc, argv );
	}

	return TCL_OK;
}

static int DioCmd( 
		  ClientData clientdata,
		  Tcl_Interp *interp,
		  int argc,
		  char **argv
		  )
{
	Device_t	*device = (Device_t *)clientdata;
	char		*dbuf;
	int		num;

	if (*++argv == NULL) {
		Tcl_SetResult( interp, "No cmd - should be one of: read, write, configure, cget or destroy", TCL_STATIC );
		return TCL_ERROR;
	}

	if (!strcmp( *argv, "read" )) {
		/* Get the number of bytes to read */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "# of bytes not specified:",
					  "should be \"",
					  "read number_of_bytes\"",
					  NULL );
			return TCL_ERROR;
		}
		if (Tcl_GetInt( interp, argv[1], &num ) != TCL_OK) {
			return TCL_ERROR;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		/* Allocate buffer for the array to read from ADC */
		dbuf = (char *)malloc( num * sizeof( char ) );
		if (dbuf == NULL) {
			Tcl_SetResult( interp, "Out of memory", TCL_STATIC );
			return TCL_ERROR;
		}

		/* Read the data from the ADC */
		if ((num = read( device->fd, dbuf,
				 num * sizeof( char ) )) < 0) {
			Tcl_AppendResult( interp, "read()",
					  Tcl_PosixError( interp ),
					  NULL );
			free( dbuf );
			return TCL_ERROR;
		}

		Tcl_SetResult( interp, dbuf, TCL_DYNAMIC );
	} else if (!strcmp( *argv, "write" )) {
		char	sbuf[100];

		/* Get the string to write to the port */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "String to write not specified:"
					  "should be \"write string\"",
					  NULL );
			return TCL_ERROR;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		/* Write the data to the device */
		if ((num = write( device->fd, argv[1], strlen(argv[1]) )) < 0) {
			Tcl_AppendResult( interp, "write()",
					  Tcl_PosixError( interp ),
					  NULL );
			return TCL_ERROR;
		}

		sprintf( sbuf, "%d", num );
		Tcl_SetResult( interp, sbuf, TCL_VOLATILE );
	} else if (!strcmp( *argv, "setbyte" )) {
		char	sbuf[100];
		char	c;

		/* Get the string to write to the port */
		if (!argv[1]) {
			Tcl_AppendResult( interp,
					  "Value for port not specified "
					  "should be \"setbyte number\"",
					  NULL );
			return TCL_ERROR;
		}

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-2, argv+2 ) != TCL_OK)
			return TCL_ERROR;

		if (Tcl_GetInt( interp, argv[1], &num ) != TCL_OK) {
			return TCL_ERROR;
		}
		c = num & 0xff;

		/* Write the data to the device */
		if ((num = write( device->fd, &c, 1 )) < 0) {
			Tcl_AppendResult( interp, "write()",
					  Tcl_PosixError( interp ),
					  NULL );
			return TCL_ERROR;
		}

		sprintf( sbuf, "%d", num );
		Tcl_SetResult( interp, sbuf, TCL_VOLATILE );
	} else if (!strcmp( *argv, "getbyte" )) {
		char	sbuf[100];
		char	c;

		/* Process any options on the command line */
		if (ProcessOptions( interp, device, argc-1, argv+1 ) != TCL_OK)
			return TCL_ERROR;

		/* Write the data to the device */
		if ((num = read( device->fd, &c, 1 )) < 0) {
			Tcl_AppendResult( interp, "read()",
					  Tcl_PosixError( interp ),
					  NULL );
			return TCL_ERROR;
		}

		sprintf( sbuf, "0x%x", (int)c );
		Tcl_SetResult( interp, sbuf, TCL_VOLATILE );
	} else {
		return ProcessCommand( interp, device, argc, argv );
	}

	return TCL_OK;
}

static void DeleteDevice( ClientData clientdata )
{
	Device_t	*device = (Device_t *)clientdata;

	if (device->handler)
		Tk_DeleteFileHandler( (ClientData) device->fd );
	if (device->filename)
		free( device->filename );
	if (device->name)
		free( device->name );
	if (device->options)
		free( device->options );
	if (device->fd >= 0)
		close( device->fd );
	free( device );
}

static int LabpcCmd(
		    ClientData clientdata,
		    Tcl_Interp *interp,
		    int argc,
		    char **argv
		    )
{
	static int	adcid = 0;
	static int	dacid = 0;
	static int	dioid = 0;
	static int	ctrid = 0;
	char		c;
	Device_t 	*device;
	char		*devtype;	/* Name for device type		*/
	char		fname[1000];	/* Dev file, eg. /dev/labpc/adc	*/
	char		cmdname[100];	/* Tcl command to create	*/
	Tcl_CmdProc	*comproc;

	if (*++argv == NULL) {
		Tcl_SetResult( interp, "Wrong # args: should be \"labpc cmd ?arg arg...?\"", TCL_STATIC );
		return TCL_ERROR;
	}

	c = **argv;
	if (!strcmp( *argv, "open" )) {
		/* Open a device - allocate the ClientData record */
		device = (Device_t *)malloc( sizeof( Device_t ) );
		if (device == NULL) {
			Tcl_SetResult( interp, "Out of memory", TCL_STATIC );
			return TCL_ERROR;
		}
		device->filename = NULL;
		device->name = NULL;
		device->fd = -1;
		device->handler = NULL;
		device->handler_interp = NULL;
		device->options = NULL;

		/* Get the device type from the command line - and
		 * load the default device filename */
		if (!(devtype = *++argv)) {
			Tcl_SetResult( interp, "Wrong # args: should be \"labpc open ?adc/dac/port/counter? ?-device pathname?\"", TCL_STATIC );
			return TCL_ERROR;
		}
		strcat( strcpy( fname, device_dir ), "/" );
		c = *devtype;
		if (c == 'a' && !strcmp( devtype, "adc" )) {
			strcat( fname, "adc" );
			sprintf( cmdname, "%s%d", "adc", adcid++ );
			device->options = (Option_t *)malloc( sizeof( AdcOptions ) );
			memcpy( device->options, AdcOptions, sizeof( AdcOptions ) );
			comproc = AdcCmd;
		} else if (c == 'c' && !strcmp( devtype, "counter" )) {
			strcat( fname, "countera0" );
			sprintf( cmdname, "%s%d", "ctr", ctrid++ );
			device->options = (Option_t *)malloc( sizeof( CtrOptions ) );
			memcpy( device->options, CtrOptions, sizeof( CtrOptions ) );
			comproc = CtrCmd;
		} else if (c == 'd' && !strcmp( devtype, "dac" )) {
			strcat( fname, "dac0" );
			sprintf( cmdname, "%s%d", "dac", dacid++ );
			device->options = (Option_t *)malloc( sizeof( DacOptions ) );
			memcpy( device->options, DacOptions, sizeof( DacOptions ) );
			comproc = DacCmd;
		} else if (c == 'd' && !strcmp( devtype, "dio" )) {
			strcat( fname, "porta" );
			sprintf( cmdname, "%s%d", "dio", dioid++ );
			device->options = (Option_t *)malloc( sizeof( DioOptions ) );
			memcpy( device->options, DioOptions, sizeof( DioOptions ) );
			comproc = DioCmd;
		} else {
			Tcl_AppendResult( interp, "Unknown device type \"",
					  *argv, "\": should be one of: adc, dac, dio or counter",
					  NULL );
			free( device );
			return TCL_ERROR;
		}

		/* Process options supplied on the command line */
		while (*++argv) {
			c = argv[0][1];
			if (c == 'd' && !strcmp( *argv, "-device" )) {
				/* Get the device filename to open */
				if (!*++argv) {
					Tcl_SetResult( interp, "No filename supplied for \"-device\" option", TCL_STATIC );
					free( device );
					return TCL_ERROR;
				}
				device->filename = strdup( *argv );
			} else {
				Tcl_AppendResult( interp, "Unknown option: ",
						  *argv, NULL );
				free( device );
				return TCL_ERROR;
			}
		}

		/* Load default parameters if not specified on command line */
		if (device->filename == NULL)
			device->filename = strdup( fname );

		/* Open the device file */
		if ((device->fd = open( device->filename, O_RDWR )) < 0) {
			Tcl_AppendResult( interp, device->filename, ": ",
					  Tcl_PosixError( interp ), NULL );
			free( device->filename );
			free( device );
			return TCL_ERROR;
		}

		device->name = strdup( cmdname );
		Tcl_SetResult( interp, cmdname, TCL_VOLATILE );
		Tcl_CreateCommand( interp, cmdname, comproc,
				   (ClientData) device, DeleteDevice );
		return TCL_OK;
	}

	Tcl_SetResult( interp, "Unknown cmd: should be one of: open, cget or configure", TCL_STATIC );
	return TCL_ERROR;
}

int Labpc_Init( Tcl_Interp *interp )
{
	Tcl_CreateCommand( interp, "labpc", LabpcCmd, (ClientData) 0,
			   (Tcl_CmdDeleteProc *) NULL);
	return TCL_OK;
}

