# Utility procs and widgets for the LabPC driver

# Tie together an entry and a label widget
proc labelentry {w name variable} {
    frame $w

    label $w.label \
	    -text "$name"
    entry $w.entry \
	    -width 6 \
	    -textvariable "$variable"
    bind $w.entry <Any-ButtonPress-3> "popuphelp $variable"

    pack append $w \
	    $w.label {left expand} \
	    $w.entry {right expand}
}

# A widget for displaying context sensitive help
proc popuphelp {name} {
    if [regexp {device([a-zA-Z]*)([0-9]*)\((.*)\)} "$name" {} device devnum option] {
	set helpname "$device -$option"
    } else {
	set helpname "$name"
    }

    set w .help_[join [split $helpname] "_"]
    catch "destroy $w"
    toplevel $w -class Help
    wm title $w "Help $helpname"

    frame $w.top
    text $w.top.text \
	    -setgrid yes \
	    -yscrollcommand "$w.top.scroll set"
    scrollbar $w.top.scroll \
	    -command "$w.top.text yview"
    button $w.button \
	    -text {Done} \
	    -command "destroy $w"

    pack append $w \
	    $w.top {top expand fill} \
	    $w.button {bottom}

    pack append $w.top \
	    $w.top.scroll {right filly} \
	    $w.top.text {left expand fill}

    # Get the helpfile name from the option database
    global labpc_dir
    set helpfile [option get $w helpfile Helpfile]
    if {"$helpfile" == ""} {
	tkerror "No helpfile set in option database"
    }

    # Open the helpfile and search for the requested entry
    set fd [open "$helpfile" "r"]
    while {[gets $fd line] >= 0} {
	if [string match "### $helpname*" "$line"] {
	    break
	}
    }

    set linknum 0
    set maxlen 20
    # Scan through the helpfile, inserting strings into the text widget
    while {[gets $fd line] >= 0} {
	if [regexp "^###" "$line"] {
	    break
	}
	set len 0
	while {[regexp {^(.*)\\link"(.*)"(.*)$} "$line" arg0 arg1 arg2 arg3]} {
	    # We've found a hyper-link make a tag for binding
	    $w.top.text insert end "$arg1" "" "$arg2" "link$linknum"
	    $w.top.text tag configure link$linknum -foreground mediumblue
	    $w.top.text tag bind link$linknum <Any-ButtonPress-1> "popuphelp \"$arg2\""
	    incr linknum
	    incr len [expr [string length "$arg1"] + [string length "$arg2"]]
	    set line "$arg3"
	}
	$w.top.text insert end "$line\n"
	incr len [string length "$line"]
	if {$len > $maxlen} {
	    set maxlen $len
	}
    }
    close $fd

    if {$maxlen > 80} {
	set maxlen 80
    }
    $w.top.text configure -state disabled -width [expr $maxlen + 1]
}

# A general purpose dialog box for entering new values for variables
proc dialogvalue {w title text arr name} {
    global $arr
    set variable ${arr}($name)

    toplevel $w -class Dialog
    wm title $w $title
    wm iconname $w Dialog

    # The message and entry widget for the top of the dialog
    frame $w.top
    label $w.top.msg -text $text
    entry $w.top.entry -width 10
    $w.top.entry delete 0 end
    $w.top.entry insert 0 [set $variable]

    # The buttons for the bottom of the dialog
    frame $w.bot
    button $w.bot.ok \
	    -text {OK} \
	    -command "set $variable \[$w.top.entry get\]; destroy $w"
    button $w.bot.clear \
	    -text {Clear} \
	    -command "$w.top.entry delete 0 end"
    button $w.bot.help \
	    -text {Help} \
	    -command "popuphelp $variable"
    button $w.bot.cancel \
	    -text {Cancel} \
	    -command "destroy $w"

    bind $w <Return> "$w.bot.ok invoke"

    pack append $w.top \
	    $w.top.msg   {top frame center fillx filly expand padx 10 pady 10} \
	    $w.top.entry {top frame center expand padx 10 pady 10}

    pack append $w.bot \
	    $w.bot.ok	  {left frame center expand pady 10 padx 10} \
	    $w.bot.clear  {left frame center expand pady 10 padx 10} \
	    $w.bot.help   {left frame center expand pady 10 padx 10} \
	    $w.bot.cancel {left frame center expand pady 10 padx 10}
	    
    pack append $w \
	    $w.top {top frame center fillx filly expand} \
	    $w.bot {top frame center fillx}
}

# A sample bar widget - can't use these yet - need select for labpc
proc createbarmeter {w} {
    canvas $w \
	    -relief sunken \
	    -height 5m \
	    -width 20m \
	    -scrollregion {0 0 40m 5m} \
	    -background LightBlue \
	    -borderwidth 2
    $w create rectangle 0 0 20m 5m \
	    -fill MediumBlue
    $w xview moveto 0.5
}

proc setbarmeter {w value} {
    $w xview moveto [expr (1.0 - $value) / 2.0]
}

proc dobartimer {w tot t dt} {
    set t [expr $t + $dt]

    if {$t >= $tot} {
	setbarmeter $w 1.0
    } else {
	setbarmeter $w [expr $t / $tot]
	after $dt dobartimer $w $tot $t $dt
    }
}

proc startbartimer {w t} {
    $w xview moveto 0.0
    dobartimer $w $t 0.0 100
}

# A simple ComboBox widget
proc combo {w arr var name values} {
    global $arr

    frame $w \
	    -relief groove \
	    -borderwidth 2

    label $w.label \
	    -text "$name"

    button $w.label2 \
	    -relief sunken \
	    -width 10 \
	    -textvariable $var
    
    bind $w.label2 <Any-ButtonPress-1> "tk_popup $w.label2.menu \[expr \[winfo rootx $w.label2\] + \[winfo width $w.label2\] / 2\] \[expr \[winfo rooty $w.label2\] + \[winfo height $w.label2\] / 2\] \[lsearch -exact {$values} \[set $var\]\]"
    bind $w.label2 <Any-ButtonPress-3> "popuphelp $var"

    set m $w.label2.menu
    menu $m -tearoff 0
    foreach opt $values {
	$m add command \
		-label "[format %-12s $opt]" \
		-command "set $var $opt"
    }

    pack append $w \
	    $w.label {left} \
	    $w.label2 {right}
}

